from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00138(RuleTemplate):
    """
    DDF00138: Every identifier must be unique within the scope of an identified organization.

    Applies to: StudyIdentifier
    Attributes: studyIdentifier
    """

    def __init__(self):
        super().__init__(
            "DDF00138",
            RuleTemplate.ERROR,
            "Every identifier must be unique within the scope of an identified organization.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0263(RuleTemplate):
    """
    CHK0263: Within a study design, if more characteristics are defined, they must be distinct.

    Applies to: InterventionalStudyDesign, ObservationalStudyDesign
    Attributes: characteristics
    """

    def __init__(self):
        super().__init__(
            "CHK0263",
            RuleTemplate.ERROR,
            "Within a study design, if more characteristics are defined, they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from .timing import Timing
from .duration import Duration


class ScheduleTimeline(ApiBaseModelWithIdNameLabelAndDesc):
    mainTimeline: bool
    entryCondition: str
    entryId: str
    exits: List[ScheduleTimelineExit] = []
    timings: List[Timing] = []
    instances: List[Union[ScheduledActivityInstance, ScheduledDecisionInstance]] = []
    plannedDuration: Union[Duration, None] = None
    instanceType: Literal["ScheduleTimeline"]

    def first_timepoint(self) -> ScheduledActivityInstance:
        return self.instances[0] if self.instances else None

    def find_timepoint(self, id: str) -> ScheduledActivityInstance:
        return next((x for x in self.instances if x.id == id), None)

    def timepoint_list(self) -> list:
        return self.instances

    def find_timing_from(self, id: str) -> Timing:
        return next(
            (x for x in self.timings if x.relativeFromScheduledInstanceId == id), None
        )

    # # @todo Nice to get StudyDesign type hint included but circular ...
    # def soa(self, study_design) -> list:
    #     # Activities
    #     activity_order = study_design.activity_list()

    #     # Epochs and Visits
    #     ai = []
    #     timepoints = self.timepoint_list()
    #     for timepoint in timepoints:
    #         timing = self.find_timing_from(timepoint.id)
    #         encounter = study_design.find_encounter(timepoint.encounterId)
    #         epoch = study_design.find_epoch(timepoint.epochId)
    #         # print(f"TPT: {timepoint}")
    #         # print(f"ENCOUNTER: {encounter}")
    #         # print(f"EPOCH: {epoch}")
    #         # print(f"TIMING: {timing}")
    #         entry = {
    #             "instance": timepoint,
    #             "timing": timing,
    #             "epoch": epoch,
    #             "encounter": encounter,
    #         }
    #         ai.append(entry)

    #     # Blank row
    #     visit_row = {}
    #     for item in ai:
    #         visit_row[item["instance"].id] = ""

    #     # Activities
    #     activities = {}
    #     for activity in activity_order:
    #         for timepoint in timepoints:
    #             if activity.id in timepoint.activityIds:
    #                 if activity.name not in activities:
    #                     activities[activity.name] = visit_row.copy()
    #                 activities[activity.name][timepoint.id] = "X"

    #     # Form results
    #     labels = []
    #     for item in ai:
    #         label = (
    #             item["encounter"].label if item["encounter"] else item["instance"].label
    #         )
    #         labels.append(label)
    #     results = []
    #     results.append(
    #         [""] + [item["epoch"].label if item["epoch"] else "&nbsp;" for item in ai]
    #     )
    #     results.append([""] + labels)
    #     results.append(
    #         [""]
    #         + [item["instance"].label if item["instance"] else "&nbsp;" for item in ai]
    #     )
    #     results.append(
    #         [""]
    #         + [
    #             item["timing"].windowLabel if item["timing"] else "&nbsp;"
    #             for item in ai
    #         ]
    #     )
    #     for activity in activity_order:
    #         if activity.name in activities:
    #             data = activities[activity.name]
    #             label = activity.label if activity.label else activity.name
    #             results.append(
    #                 [{"name": label, "id": activity.id}] + list(data.values())
    #             )

    #     # Format as HTML
    #     lines = []
    #     lines.append('<table class="soa-table table">')
    #     lines.append("<thead>")
    #     lines.append('<tr class="table-active">')
    #     epochs = [[i, len([*group])] for i, group in groupby(results[0])]
    #     text = '<td><p class="m-0 p-0"><small>&nbsp;</small></p></td>'
    #     for epoch in epochs[1:]:
    #         text += f'<td class="text-center" colspan="{epoch[1]}"><p class="m-0 p-0"><small>{epoch[0]}</small></p></td>'
    #     lines.append(text)
    #     lines.append("</tr>")
    #     for result in results[1:4]:
    #         lines.append('<tr class="table-active">')
    #         lines.append('<td><p class="m-0 p-0"><small>&nbsp;</small></p></td>')
    #         for cell in result[1:]:
    #             lines.append(
    #                 f'<td class="text-center"><p class="m-0 p-0"><small>{cell}</small></p></td>'
    #             )
    #         lines.append("</tr>")
    #     lines.append("</thead>")
    #     lines.append("<tbody>")
    #     for result in results[4:]:
    #         lines.append("<tr>")
    #         x = result[0]["name"]
    #         lines.append(
    #             f'<td class="m-0 p-0"><p class="m-0 p-0"><small>{x}</small></p></td>'
    #         )
    #         for cell in result[1:]:
    #             lines.append(
    #                 f'<td class="m-0 p-0 text-center"><p class="m-0 p-0"><small>{cell}</small></p></td>'
    #             )
    #         lines.append("</tr>")
    #     lines.append("</tbody>")
    #     lines.append("</table>")
    #     return ("\n").join(lines)



from datetime import date
from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .identifier import StudyIdentifier, ReferenceIdentifier
from .study_design import (
    StudyDesign,
    InterventionalStudyDesign,
    ObservationalStudyDesign,
)
from .governance_date import GovernanceDate
from .study_amendment import StudyAmendment
from .study_title import StudyTitle
from .eligibility_criterion import EligibilityCriterionItem
from .narrative_content import NarrativeContentItem
from .comment_annotation import CommentAnnotation
from .abbreviation import Abbreviation
from .study_role import StudyRole
from .organization import Organization
from .study_intervention import StudyIntervention
from .administrable_product import AdministrableProduct
from .medical_device import MedicalDevice
from .product_organization_role import ProductOrganizationRole
from .biomedical_concept import BiomedicalConcept
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .syntax_template_dictionary import SyntaxTemplateDictionary
from .condition import Condition
from .extension import Extension


CS_EXT_URL = "www.d4k.dk/usdm/extensions/001"

class StudyVersion(ApiBaseModelWithId):
    versionIdentifier: str
    rationale: str
    documentVersionIds: List[str] = []
    dateValues: List[GovernanceDate] = []
    amendments: List[StudyAmendment] = []
    businessTherapeuticAreas: List[Code] = []
    studyIdentifiers: List[StudyIdentifier]
    referenceIdentifiers: List[ReferenceIdentifier] = []
    studyDesigns: List[Union[InterventionalStudyDesign, ObservationalStudyDesign]] = []
    titles: List[StudyTitle]
    eligibilityCriterionItems: List[EligibilityCriterionItem] = []
    narrativeContentItems: List[NarrativeContentItem] = []
    abbreviations: List[Abbreviation] = []
    roles: List[StudyRole] = []
    organizations: List[Organization] = []
    studyInterventions: List[StudyIntervention] = []
    administrableProducts: List[AdministrableProduct] = []
    medicalDevices: List[MedicalDevice] = []
    productOrganizationRoles: List[ProductOrganizationRole] = []
    biomedicalConcepts: List[BiomedicalConcept] = []
    bcCategories: List[BiomedicalConceptCategory] = []
    bcSurrogates: List[BiomedicalConceptSurrogate] = []
    dictionaries: List[SyntaxTemplateDictionary] = []
    conditions: List[Condition] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyVersion"]

    def confidentiality_statement(self):
        ext: Extension = self.get_extension(CS_EXT_URL)
        return ext.valueString if ext else ""

    def get_title(self, title_type):
        for title in self.titles:
            if title.type.decode == title_type:
                return title
        return None

    def sponsor_identifier(self) -> StudyIdentifier | None:
        for identifier in self.studyIdentifiers:
            org = self.organization(identifier.scopeId)
            if org and org.type.code == "C54149":
                return identifier
        return None

    def regulatory_identifiers(self) -> list[StudyIdentifier]:
        results = []
        for identifier in self.studyIdentifiers:
            org = self.organization(identifier.scopeId)
            if org and org.type.code == "C188863":
                results.append(identifier)
        return results

    def registry_identifiers(self) -> list[StudyIdentifier]:
        results = []
        for identifier in self.studyIdentifiers:
            org = self.organization(identifier.scopeId)
            if org and org.type.code == "C93453":
                results.append(identifier)
        return results

    def organization(self, id: str) -> Organization:
        return next((x for x in self.organizations if x.id == id), None)

    def criterion_item(self, id: str) -> EligibilityCriterionItem:
        return next((x for x in self.eligibilityCriterionItems if x.id == id), None)

    def intervention(self, id: str) -> StudyIntervention:
        return next((x for x in self.studyInterventions if x.id == id), None)

    def condition(self, id: str) -> Condition:
        return next((x for x in self.conditions if x.id == id), None)

    def phases(self) -> str:
        return ", ".join([sd.phase_as_text() for sd in self.studyDesigns])

    def official_title_text(self) -> str:
        for x in self.titles:
            if x.is_official():
                return x.text
        return ""

    def short_title_text(self) -> str:
        for x in self.titles:
            if x.is_short():
                return x.text
        return ""

    def acronym_text(self) -> str:
        for x in self.titles:
            if x.is_acronym():
                return x.text
        return ""

    def official_title(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_official():
                return x
        return None

    def short_title(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_short():
                return x
        return None

    def acronym(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_acronym():
                return x
        return None

    def sponsor(self) -> Organization:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if x.is_sponsor(map):
                return map[x.scopeId]
        return None

    # Note: Method sponsor_identifier in base USDM class

    def sponsor_identifier_text(self) -> StudyIdentifier:
        for x in self.studyIdentifiers:
            if x.is_sponsor(self.organization_map()):
                return x.text
        return ""

    def sponsor_name(self) -> str:
        map = self.organization_map()
        x: StudyIdentifier
        for x in self.studyIdentifiers:
            if x.is_sponsor(map):
                return map[x.scopeId].name
        return ""

    def sponsor_label(self) -> str:
        map: dict[Organization] = self.organization_map()
        x: StudyIdentifier
        for x in self.studyIdentifiers:
            if x.is_sponsor(map):
                return map[x.scopeId].label
        return ""

    def sponsor_label_name(self) -> str:
        label = self.sponsor_label()
        return label if label else self.sponsor_name()

    def sponsor_address(self) -> str:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if x.is_sponsor(map):
                return (
                    map[x.scopeId].legalAddress.text
                    if map[x.scopeId].legalAddress
                    else ""
                )
        return ""

    def nct_identifier(self) -> StudyIdentifier:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if map[x.scopeId].name == "ClinicalTrials.gov":
                return x.text
        return ""

    def protocol_date(self) -> GovernanceDate:
        for x in self.dateValues:
            if x.type.decode == "Protocol Effective Date":
                return x
        return ""

    def approval_date(self) -> GovernanceDate:
        for x in self.dateValues:
            if x.type.decode == "Protocol Approval by Sponsor Date":
                return x
        return ""

    def protocol_date_value(self) -> date:
        for x in self.dateValues:
            if x.type.decode == "Protocol Effective Date":
                return x.dateValue
        return ""

    def approval_date_value(self) -> date:
        for x in self.dateValues:
            if x.type.decode == "Protocol Approval by Sponsor Date":
                return x.dateValue
        return ""

    def organization_map(self) -> Organization:
        return {x.id: x for x in self.organizations}

    def find_study_design(self, id: str) -> StudyDesign:
        return next((x for x in self.studyDesigns if x.id == id), None)

    def narrative_content_item_map(self) -> NarrativeContentItem:
        return {x.id: x for x in self.narrativeContentItems}



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0268(RuleTemplate):
    """
    CHK0268: An observational study design's time perspective must be specified according to the extensible Observational Study Time Perspective (C127261) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: ObservationalStudyDesign
    Attributes: timePerspective
    """

    def __init__(self):
        super().__init__(
            "CHK0268",
            RuleTemplate.ERROR,
            "An observational study design's time perspective must be specified according to the extensible Observational Study Time Perspective (C127261) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00071(RuleTemplate):
    """
    DDF00071: A study cell must only reference an arm that is defined within the same study design as the study cell.

    Applies to: StudyCell
    Attributes: arm
    """

    def __init__(self):
        super().__init__(
            "DDF00071",
            RuleTemplate.ERROR,
            "A study cell must only reference an arm that is defined within the same study design as the study cell.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from datetime import date
from .code import Code
from .geographic_scope import GeographicScope


class GovernanceDate(ApiBaseModelWithIdNameLabelAndDesc):
    type: Code
    dateValue: date
    geographicScopes: List[GeographicScope]
    instanceType: Literal["GovernanceDate"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00069(RuleTemplate):
    """
    DDF00069: Each combination of arm and epoch must occur no more than once within a study design.

    Applies to: StudyCell
    Attributes: arm, epoch
    """

    def __init__(self):
        super().__init__(
            "DDF00069",
            RuleTemplate.ERROR,
            "Each combination of arm and epoch must occur no more than once within a study design.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00044(RuleTemplate):
    """
    DDF00044: The target for a condition must not be equal to its parent.

    Applies to: ConditionAssignment
    Attributes: conditionTarget
    """

    def __init__(self):
        super().__init__(
            "DDF00044",
            RuleTemplate.ERROR,
            "The target for a condition must not be equal to its parent.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .masking import Masking
from .assigned_person import AssignedPerson
from .comment_annotation import CommentAnnotation


class StudyRole(ApiBaseModelWithIdNameLabelAndDesc):
    code: Code
    appliesToIds: List[str] = []  # Going to allow for an empty list but really a role
    # should be attached to a Study Version or Design
    assignedPersons: List[AssignedPerson] = []
    organizationIds: List[str] = []
    masking: Union[Masking, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyRole"]



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .alias_code import AliasCode


class QuantityRange(ApiBaseModelWithId):
    pass


class Quantity(QuantityRange):
    value: float
    unit: Union[AliasCode, None] = None
    instanceType: Literal["Quantity"]


class Range(QuantityRange):
    minValue: Quantity
    maxValue: Quantity
    isApproximate: bool
    instanceType: Literal["Range"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00070(RuleTemplate):
    """
    DDF00070: The minimum value of a range must be less than or equal to the maximum value of the range.

    Applies to: Range
    Attributes: minValue, maxValue
    """

    def __init__(self):
        super().__init__(
            "DDF00070",
            RuleTemplate.ERROR,
            "The minimum value of a range must be less than or equal to the maximum value of the range.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.builder.builder import Builder
from usdm4.api.narrative_content import NarrativeContent, NarrativeContentItem
from usdm4.api.study_definition_document import StudyDefinitionDocument
from usdm4.api.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm4.api.geographic_scope import GeographicScope
from usdm4.api.governance_date import GovernanceDate
from usdm4.assembler.encoder import Encoder


class DocumentAssembler(BaseAssembler):
    """
    Assembler responsible for creating StudyDefinitionDocument and StudyDefinitionDocumentVersion objects
    from document data and structured content sections.

    This assembler processes protocol documents, their versions, and hierarchical content sections,
    creating narrative content structures with proper cross-references and governance dates.
    It handles the conversion of structured document data into USDM-compliant document objects.
    """

    MODULE = "usdm4.assembler.document_assembler.DocumentAssembler"
    DIV_OPEN_NS = '<div xmlns="http://www.w3.org/1999/xhtml">'
    DIV_CLOSE = "</div>"

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the DocumentAssembler.

        Args:
            builder (Builder): The builder instance for creating USDM objects
            errors (Errors): Error handling instance for logging issues
        """
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        self._document: StudyDefinitionDocument = None
        self._document_version: StudyDefinitionDocumentVersion = None
        self._contents = []
        self._dates = []

    def execute(self, data: dict) -> None:
        """
        Creates StudyDefinitionDocument and StudyDefinitionDocumentVersion objects from document data.

        Args:
            data (dict): A dictionary containing document and section information.
                        The data parameter must have the following structure:

                        {
                            "document": {               # Document metadata
                                "label": str,           # Human-readable document label
                                "version": str,         # Document version identifier
                                "status": str,          # Document status (e.g., "Draft", "Final")
                                "template": str,        # Template name used for document
                                "version_date": str,    # ISO format date for document version
                            },
                            "sections": [               # List of document sections
                                {
                                    "section_number": str,    # Section number (e.g., "1.1", "2.3.1")
                                    "section_title": str,     # Section title/heading
                                    "text": str,              # Section content (HTML allowed)
                                },
                                # ... additional sections
                            ]
                        }

                        Required fields in "document":
                        - "label": Display name for the document
                        - "version": Version identifier (e.g., "1.0", "2.1")
                        - "status": Current status of the document
                        - "template": Template used for document generation
                        - "version_date": Date when this version was created (ISO format)

                        Required fields in each "sections" item:
                        - "section_number": Hierarchical section number (determines nesting level)
                        - "section_title": Title/heading for the section
                        - "text": Content of the section (can include HTML markup)

                        Note: Section hierarchy is determined by the depth of the section_number
                        (e.g., "1" = level 1, "1.1" = level 2, "1.1.1" = level 3)

        Returns:
            None: Created objects are stored in instance properties

        Note:
            This method creates both document metadata objects and processes hierarchical
            sections into NarrativeContent structures with proper parent-child relationships
            and sequential linking (previousId/nextId).
        """
        try:
            if data:
                # Extract document metadata and sections from input data
                document: dict = data["document"]
                sections: list[dict] = data["sections"]

                # Create governance date from document version date
                self._create_date(document)

                # Create document version object with version and status
                self._document_version = self._builder.create(
                    StudyDefinitionDocumentVersion,
                    {
                        "version": document["version"],  # Version identifier from input
                        "status": self._encoder.document_status(
                            document["status"]
                        ),  # Document status from input
                    },
                )

                # Get standard codes for document properties
                language = self._builder.iso639_code("en")  # Default to English
                doc_type = self._builder.cdisc_code(
                    "C70817", "Protocol"
                )  # CDISC code for Protocol

                # Create the main document object
                self._document = self._builder.create(
                    StudyDefinitionDocument,
                    {
                        "name": self._label_to_name(
                            document["label"]
                        ),  # Convert label to internal name
                        "label": document["label"],  # Display label from input
                        "description": "Protocol Document",  # Default description
                        "language": language,  # ISO language code
                        "type": doc_type,  # CDISC document type code
                        "templateName": document[
                            "template"
                        ],  # Template name from input
                        "versions": [
                            self._document_version
                        ],  # Include the created version
                    },
                )

                # Process sections into hierarchical narrative content structure
                _ = self._section_to_narrative(None, sections, 0, 1)

                # Create sequential links between narrative content items
                self._builder.double_link(
                    self._document_version.contents, "previousId", "nextId"
                )
            else:
                self._errors.info(
                    "No document to build, no data",
                    KlassMethodLocation(self.MODULE, "execute"),
                )
        except Exception as e:
            self._errors.exception(
                "Failed during creation of document",
                e,
                KlassMethodLocation(self.MODULE, "execute"),
            )

    @property
    def document(self) -> StudyDefinitionDocument:
        return self._document

    @property
    def document_version(self) -> StudyDefinitionDocumentVersion:
        return self._document_version

    @property
    def contents(self) -> list[NarrativeContentItem]:
        # print(f"SECTION COUNT: {len(self._contents)}")
        return self._contents

    @property
    def dates(self) -> list[GovernanceDate]:
        return self._dates

    def _section_to_narrative(
        self, parent: NarrativeContent, sections: list[dict], index: int, level: int
    ) -> int:
        process = True
        previous = None
        local_index = index
        location = KlassMethodLocation(self.MODULE, "_section_to_narrative")
        while process:
            section = sections[local_index]
            section_level = self._section_level(section)
            if section_level == level:
                sn = section["section_number"] if section["section_number"] else ""
                dsn = True if sn else False
                st = section["section_title"] if section["section_title"] else ""
                dst = True if st else False
                nc_text = f"{self.DIV_OPEN_NS}{section['text']}{self.DIV_CLOSE}"
                nci = self._builder.create(
                    NarrativeContentItem,
                    {"name": f"NCI-{local_index}", "text": nc_text},
                )
                nc = self._builder.create(
                    NarrativeContent,
                    {
                        "name": f"NC-{local_index}",
                        "sectionNumber": sn,
                        "displaySectionNumber": dsn,
                        "sectionTitle": st,
                        "displaySectionTitle": dst,
                        "contentItemId": nci.id,
                        "childIds": [],
                        "previousId": None,
                        "nextId": None,
                    },
                )
                self._document_version.contents.append(nc)
                self._contents.append(nci)
                # print(f"SECTION ADD: {len(self._contents)}")
                if parent:
                    parent.childIds.append(nc.id)
                previous = nc
                local_index += 1
            elif section_level > level:
                if previous:
                    local_index = self._section_to_narrative(
                        previous, sections, local_index, level + 1
                    )
                else:
                    self._errors.error(
                        "No previous section set while processing sections", location
                    )
                    local_index += 1
            elif section_level < level:
                return local_index
            if local_index >= len(sections):
                process = False
        return local_index

    def _section_level(self, section: dict) -> int:
        section_number: str = section["section_number"]
        text = section_number[:-1] if section_number.endswith(".") else section_number
        return len(text.split("."))

    def _create_date(self, data: dict) -> None:
        """
        Creates a governance date for the protocol effective date from document data.

        Args:
            data (dict): A dictionary containing document date information.
                        Expected structure for date creation:

                        {
                            "version_date": str,       # ISO format date string (YYYY-MM-DD)
                            # Additional fields from document metadata:
                            # "label": str,            # Document label
                            # "version": str,          # Document version
                            # "status": str,           # Document status
                            # "template": str,         # Template name
                        }

                        Required fields:
                        - "version_date": Date when the protocol version became effective
                          (must be in ISO date format)

        Returns:
            None: Created date is stored in self._dates list

        Note:
            This method creates a GovernanceDate object with CDISC-compliant codes for
            protocol effective date type and global geographic scope. The created date
            is later combined with other dates in the study assembler.
        """
        try:
            if actual_date := self._encoder.to_date(data["version_date"]):
                protocol_date_code = self._builder.cdisc_code(
                    "C207598",
                    "Protocol Effective Date",
                )
                global_code = self._builder.cdisc_code("C68846", "Global")
                global_scope = self._builder.create(
                    GeographicScope, {"type": global_code}
                )
                protocol_date = self._builder.create(
                    GovernanceDate,
                    {
                        "name": "PROTOCOL-DATE",  # Internal name for the date
                        "type": protocol_date_code,  # CDISC code for date type
                        "dateValue": actual_date,
                        "geographicScopes": [global_scope],  # Global scope application
                    },
                )
                if protocol_date:
                    self._dates.append(protocol_date)
            else:
                self._errors.warning(
                    "No sponsor approval date detected",
                    KlassMethodLocation(self.MODULE, "_create_date"),
                )
        except Exception as e:
            self._errors.exception(
                "Failed during creation of governance date",
                e,
                KlassMethodLocation(self.MODULE, "_create_date"),
            )



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00004(RuleTemplate):
    """
    DDF00004: If duration will vary (attribute durationWillVary is True) then a reason (attribute reasonDurationWillVary) must be given and vice versa.

    Applies to: AdministrationDuration
    Attributes: durationWillVary, reasonDurationWillVary
    """

    def __init__(self):
        super().__init__(
            "DDF00004",
            RuleTemplate.ERROR,
            "If duration will vary (attribute durationWillVary is True) then a reason (attribute reasonDurationWillVary) must be given and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.builder.builder import Builder
from usdm4.assembler.encoder import Encoder
from usdm4.api.schedule_timeline import ScheduleTimeline
from usdm4.api.schedule_timeline_exit import ScheduleTimelineExit
from usdm4.api.scheduled_instance import ScheduledInstance, ScheduledActivityInstance
from usdm4.api.activity import Activity
from usdm4.api.study_epoch import StudyEpoch
from usdm4.api.encounter import Encounter
from usdm4.api.timing import Timing
from usdm4.api.condition import Condition


class TimelineAssembler(BaseAssembler):
    MODULE = "usdm4.assembler.timeline_assembler.TimelineAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        self._timelines: list[ScheduleTimeline] = []
        self._epochs: list[StudyEpoch] = []
        self._encounters: list[Encounter] = []
        self._activities: list[Activity] = []
        self._condition_links: dict = {}
        self._conditions = []

    def execute(self, data: dict) -> None:
        try:
            self._epochs = self._add_epochs(data)
            self._encounters = self._add_encounters(data)
            self._activities = self._add_activities(data)
            timepoints = self._add_timepoints(data)
            timings = self._add_timing(data)
            self._link_timepoints_and_activities(data)
            self._conditions = self._add_conditions(data)
            tl = self._add_timeline(data, timepoints, timings)
            self._timelines.append(tl)
        except Exception as e:
            self._errors.exception(
                "Failed during creation of study design",
                e,
                KlassMethodLocation(self.MODULE, "execute"),
            )

    @property
    def timelines(self) -> list[ScheduleTimeline]:
        return self._timelines

    @property
    def encounters(self) -> list[Encounter]:
        return self._encounters

    @property
    def epochs(self) -> list[StudyEpoch]:
        return self._epochs

    @property
    def activities(self) -> list[Activity]:
        return self._activities

    @property
    def conditions(self) -> list[Condition]:
        return self._conditions

    def _add_epochs(self, data) -> list[ScheduledInstance]:
        try:
            results = []
            map = {}
            self._errors.debug(
                f"EPOCHS:\n{data['epochs']}\n",
                KlassMethodLocation(self.MODULE, "_add_epochs"),
            )
            items = data["epochs"]["items"]
            timepoints = data["timepoints"]["items"]
            for index, item in enumerate(items):
                label = item["text"]
                name = f"EPOCH-{label.upper()}"
                if name not in map:
                    epoch: StudyEpoch = self._builder.create(
                        StudyEpoch,
                        {
                            "name": f"EPOCH-{index + 1}",
                            "description": f"EPOCH-{name}",
                            "label": label,
                            "type": self._builder.klass_and_attribute_value(
                                StudyEpoch, "type", "Treatment Epoch"
                            ),
                        },
                    )
                    results.append(epoch)
                    map[name] = epoch
                epoch = map[name]
                timepoints[index]["epoch_instance"] = epoch
            self._errors.info(
                f"Epochs: {len(results)}",
                KlassMethodLocation(self.MODULE, "_add_epochs"),
            )
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating Epochs",
                e,
                KlassMethodLocation(self.MODULE, "_add_epochs"),
            )
            return results

    def _add_encounters(self, data) -> list[Encounter]:
        try:
            results = []
            items = data["visits"]["items"]
            timepoints: dict = data["timepoints"]["items"]
            for index, item in enumerate(items):
                name = item["text"]
                encounter: Encounter = self._builder.create(
                    Encounter,
                    {
                        "name": f"ENCOUNTER-{index + 1}",
                        "description": f"Encounter {name}",
                        "label": name,
                        "type": self._builder.klass_and_attribute_value(
                            Encounter, "type", "visit"
                        ),
                        "environmentalSettings": [
                            self._builder.klass_and_attribute_value(
                                Encounter, "environmentalSettings", "clinic"
                            )
                        ],
                        "contactModes": [
                            self._builder.klass_and_attribute_value(
                                Encounter, "contactModes", "In Person"
                            )
                        ],
                        "transitionStartRule": None,
                        "transitionEndRule": None,
                        "scheduledAtId": None,  # @todo
                    },
                )
                results.append(encounter)
                timepoints[index]["encounter_instance"] = encounter
                for ref in item["references"]:
                    self._condition_timepoint_index(ref, index)
            self._errors.info(
                f"Encounters: {len(results)}",
                KlassMethodLocation(self.MODULE, "_add_encounters"),
            )
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating Encounters",
                e,
                KlassMethodLocation(self.MODULE, "_add_encounters"),
            )
            return results

    def _condition_timepoint_index(self, ref: str, index: int) -> None:
        if ref not in self._condition_links:
            self._condition_links[ref] = {
                "reference": ref,
                "timepoint_index": [],
                "activity_id": [],
            }
        self._condition_links[ref]["timepoint_index"].append(index)

    def _condition_activity_id(self, ref: str, id: str) -> None:
        if ref not in self._condition_links:
            self._condition_links[ref] = {
                "reference": ref,
                "timepoint_index": [],
                "activity_id": [],
            }
        self._condition_links[ref]["activity_id"].append(id)

    def _condition_combined(self, ref, sai_index: int, activity_id: str) -> None:
        if ref not in self._condition_links:
            self._condition_links[ref] = {
                "reference": ref,
                "timepoint_index": [],
                "activity_id": [],
            }
        self._condition_links[ref]["activity_id"].append(activity_id)
        self._condition_links[ref]["timepoint_index"].append(sai_index)

    def _add_activities(self, data) -> list[Activity]:
        try:
            results = []
            items = data["activities"]["items"]
            for index, item in enumerate(items):
                # print(f"ADDING ACTIVITY: {index}, {item}")
                params = {
                    "name": f"ACTIVITY-{index + 1}",
                    "description": f"Activity {item['name']}",
                    "label": item["name"],
                    "definedProcedures": [],
                    "biomedicalConceptIds": [],
                    "bcCategoryIds": [],
                    "bcSurrogateIds": [],
                    "timelineId": None,
                }
                activity: Activity = self._builder.create(Activity, params)
                results.append(activity)
                if "references" in item:
                    for ref in item["references"]:
                        # print(f"ADDING ACTIVITY REF: {item["name"]} -> {ref}")
                        self._condition_activity_id(ref, activity.id)
                item["activity_instance"] = activity
                if "children" in item:
                    for child in item["children"]:
                        # print(f"ADDING ACTIVITY: _, {child}")
                        params = {
                            "name": f"ACTIVITY-{child['name'].upper()}",
                            "description": f"Activity {child['name']}",
                            "label": child["name"],
                            "definedProcedures": [],
                            "biomedicalConceptIds": [],
                            "bcCategoryIds": [],
                            "bcSurrogateIds": [],
                            "timelineId": None,
                        }
                        child_activity: Activity = self._builder.create(
                            Activity, params
                        )
                        results.append(child_activity)
                        if "references" in child:
                            for ref in child["references"]:
                                # print(f"ADDING ACTIVITY REF: {item["name"]} -> {ref}")
                                self._condition_activity_id(ref, child_activity.id)
                        child["activity_instance"] = child_activity
                        activity.childIds.append(child_activity.id)
            self._errors.info(
                f"Activities: {len(results)}",
                KlassMethodLocation(self.MODULE, "_add_activities"),
            )
            self._builder.double_link(results, "previousId", "nextId")
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating Activities",
                e,
                KlassMethodLocation(self.MODULE, "_add_activities"),
            )
            return results

    def _add_timepoints(self, data) -> list[ScheduledInstance]:
        try:
            results = []
            timepoints: list = data["timepoints"]["items"]
            # epochs: list = data["epochs"]["items"]
            # encounters: list = data["visits"]["items"]
            for index, item in enumerate(timepoints):
                sai = self._builder.create(
                    ScheduledActivityInstance,
                    {
                        "name": f"SAI-{index + 1}",
                        "description": f"Scheduled activity instance {index + 1}",
                        "label": item["text"],
                        "timelineExitId": None,
                        "encounterId": item["encounter_instance"].id
                        if item["encounter_instance"]
                        else None,
                        "scheduledInstanceTimelineId": None,
                        "defaultConditionId": None,
                        "epochId": item["epoch_instance"].id,
                        "activityIds": [],
                    },
                )
                item["sai_instance"] = sai
                results.append(sai)
            self._errors.info(
                f"SAI: {len(results)}",
                KlassMethodLocation(self.MODULE, "_add_timepoints"),
            )
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating Scheduled Activity timepoints",
                e,
                KlassMethodLocation(self.MODULE, "_add_timepoints"),
            )
            return results

    def _add_conditions(self, data) -> list[Condition]:
        results = []
        conditions: list = data["conditions"]["items"]
        timepoints: list = data["timepoints"]["items"]
        # print(f"COND LINKS: {self._condition_links:}")
        try:
            for index, item in enumerate(conditions):
                # print(f"COND: {item}")
                if ref := item["reference"]:
                    if ref in self._condition_links:
                        # print(f"COND REF 1: {ref}")
                        links = self._condition_links[ref]
                        timepoint_ids = [
                            timepoints[x]["sai_instance"].id
                            for x in links["timepoint_index"]
                        ]
                        activity_ids = [x for x in links["activity_id"]]
                        condition = self._builder.create(
                            Condition,
                            {
                                "name": f"Condition_{index + 1}",
                                "label": f"Condition {index + 1}",
                                "description": f"Extracted footnote / condition {index + 1}",
                                "text": item["text"],
                                "dictionaryId": None,
                                "contextIds": timepoint_ids
                                if timepoint_ids
                                else activity_ids,
                                "appliesToIds": activity_ids if timepoint_ids else [],
                            },
                        )
                        if condition:
                            # print(f"COND REF 2: {condition}")
                            results.append(condition)
                    else:
                        # print(f"COND LINKS: {self._condition_links:}")
                        self._errors.warning(
                            f"Failed to align condition {item}, not created.",
                            KlassMethodLocation(self.MODULE, "_add_conditions"),
                        )
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating conditions",
                e,
                KlassMethodLocation(self.MODULE, "_add_conditions"),
            )
            return results

    def _add_timing(self, data) -> list[ScheduledInstance]:
        try:
            results = []
            timepoints: list = data["timepoints"]["items"]
            anchor_index = self._find_anchor(data)
            anchor: ScheduledInstance = timepoints[anchor_index]["sai_instance"]
            item: dict[str]
            for index, item in enumerate(timepoints):
                this_sai: ScheduledInstance = item["sai_instance"]
                if index < anchor_index:
                    if timing := self._timing(
                        data, index, "Before", this_sai.id, anchor.id
                    ):
                        results.append(timing)
                elif index == anchor_index:
                    if timing := self._timing(
                        data, index, "Fixed Reference", this_sai.id, this_sai.id
                    ):
                        results.append(timing)
                else:
                    if timing := self._timing(
                        data, index, "After", this_sai.id, anchor.id
                    ):
                        results.append(timing)
            self._errors.info(
                f"Timing: {len(results)}",
                KlassMethodLocation(self.MODULE, "_add_timing"),
            )
            return results
        except Exception as e:
            self._errors.exception(
                "Error creating timings",
                e,
                KlassMethodLocation(self.MODULE, "_add_timing"),
            )
            return results

    def _timing(
        self, data: dict, index: int, type: str, from_id: str, to_id: str
    ) -> Timing:
        try:
            windows: list = data["windows"]["items"]
            timepoints: list = data["timepoints"]["items"]
            timepoint = timepoints[index]
            window = windows[index]
            item: Timing = self._builder.create(
                Timing,
                {
                    "type": self._builder.klass_and_attribute_value(
                        Timing, "type", type
                    ),
                    "value": self._encoder.iso8601_duration(
                        timepoint["value"], timepoint["unit"]
                    ),
                    "valueLabel": self._timing_value_label(timepoints, index),
                    "name": f"TIMING-{index}",
                    "description": f"Timing {index + 1}",
                    "label": self._timing_value_label(timepoints, index),
                    "relativeToFrom": self._builder.klass_and_attribute_value(
                        Timing, "relativeToFrom", "start to start"
                    ),
                    "windowLabel": self._window_label(windows, index),
                    "windowLower": self._encoder.iso8601_duration(
                        window["before"], window["unit"]
                    )
                    if window["before"]
                    else "",
                    "windowUpper": self._encoder.iso8601_duration(
                        window["after"], window["unit"]
                    )
                    if window["after"]
                    else "",
                    "relativeFromScheduledInstanceId": from_id,
                    "relativeToScheduledInstanceId": to_id,
                },
            )
            # print(f"WINDOW: {window} -> {item.windowLabel}, [{item.windowLower}, {item.windowUpper}]")
            return item
        except Exception as e:
            self._errors.exception(
                "Error creating individual timing",
                e,
                KlassMethodLocation(self.MODULE, "_timing"),
            )
            return None

    def _window_label(self, windows: list[dict], index: int) -> str:
        if index >= len(windows):
            return "???"
        window = windows[index]
        if window["before"] == 0 and window["after"] == 0:
            return ""
        return f"-{window['before']}..+{window['after']} {window['unit']}"

    def _timing_value_label(self, timepoints: list[dict], index: int) -> str:
        if index >= len(timepoints):
            return "???"
        return f"{timepoints[index]['text']}" if timepoints[index]["text"] else "???"

    def _find_anchor(self, data) -> int:
        items = data["timepoints"]["items"]
        item: dict
        for item in items:
            if item["value"] == "1":
                item["sai_instance"]
                return int(item["index"])
        return 0

    def _link_timepoints_and_activities(self, data: dict) -> None:
        try:
            activities = data["activities"]["items"]
            timepoints = data["timepoints"]["items"]
            for a_index, activity in enumerate(activities):
                if "children" in activity:
                    for child in activity["children"]:
                        # sai_index = child["index"]
                        activity_instance: Activity = child["activity_instance"]
                        for visit in child["visits"]:
                            index = visit["index"]
                            sai_instance: ScheduledActivityInstance = timepoints[index][
                                "sai_instance"
                            ]
                            sai_instance.activityIds.append(activity_instance.id)
                            for ref in visit["references"]:
                                self._condition_combined(
                                    ref, index, activity_instance.id
                                )
                else:
                    activity_instance: Activity = activity["activity_instance"]
                    for visit in activity["visits"]:
                        index = visit["index"]
                        sai_instance: ScheduledActivityInstance = timepoints[index][
                            "sai_instance"
                        ]
                        sai_instance.activityIds.append(activity_instance.id)
                        for ref in visit["references"]:
                            self._condition_combined(ref, index, activity_instance.id)
        except Exception as e:
            self._errors.exception(
                "Error linking timepoints and activities",
                e,
                KlassMethodLocation(self.MODULE, "_link_timepoints_and_activities"),
            )
            return None

    def _add_timeline(
        self, data, instances: list[ScheduledInstance], timings: list[Timing]
    ):
        try:
            self._errors.debug(
                f"Instances: {len(instances)}, Timings: {len(timings)}",
                KlassMethodLocation(self.MODULE, "_add_timeline"),
            )
            exit = self._builder.create(ScheduleTimelineExit, {})
            duration = None
            return self._builder.create(
                ScheduleTimeline,
                {
                    "mainTimeline": True,
                    "name": "TIMELINE-1",
                    "description": "The main timeline",
                    "label": "Main timeline",
                    "entryCondition": "Paricipant identified",
                    "entryId": instances[0].id,
                    "exits": [exit],
                    "plannedDuration": duration,
                    "instances": instances,
                    "timings": timings,
                },
            )
        except Exception as e:
            self._errors.exception(
                "Error creating timeline",
                e,
                KlassMethodLocation(self.MODULE, "_add_timeline"),
            )
            return None



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00041(RuleTemplate):
    """
    DDF00041: Within a study design, there must be at least one endpoint with level primary.

    Applies to: Endpoint
    Attributes: level
    """

    def __init__(self):
        super().__init__(
            "DDF00041",
            RuleTemplate.ERROR,
            "Within a study design, there must be at least one endpoint with level primary.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00054(RuleTemplate):
    """
    DDF00054: Within an encounter there must be no duplicate contact modes.

    Applies to: Encounter
    Attributes: contactModes
    """

    def __init__(self):
        super().__init__(
            "DDF00054",
            RuleTemplate.ERROR,
            "Within an encounter there must be no duplicate contact modes.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from .abbreviation import Abbreviation
from .activity import Activity
from .address import Address
from .duration import Duration
from .administration import Administration
from .administrable_product import AdministrableProduct
from .administrable_product_property import AdministrableProductProperty
from .alias_code import AliasCode
from .analysis_population import AnalysisPopulation
from .assigned_person import AssignedPerson
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_property import BiomedicalConceptProperty
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .biomedical_concept import BiomedicalConcept
from .characteristic import Characteristic
from .code import Code
from .condition import Condition
from .document_content_reference import DocumentContentReference
from .eligibility_criterion import EligibilityCriterion, EligibilityCriterionItem
from .encounter import Encounter
from .endpoint import Endpoint
from .estimand import Estimand
from .extension import ExtensionAttribute, ExtensionClass
from .geographic_scope import GeographicScope
from .governance_date import GovernanceDate
from .indication import Indication
from .ingredient import Ingredient
from .intercurrent_event import IntercurrentEvent
from .masking import Masking
from .narrative_content import NarrativeContent, NarrativeContentItem
from .objective import Objective
from .organization import Organization
from .population_definition import StudyDesignPopulation, StudyCohort
from .procedure import Procedure
from .quantity_range import Quantity
from .quantity_range import Range
from .response_code import ResponseCode
from .schedule_timeline_exit import ScheduleTimelineExit
from .schedule_timeline import ScheduleTimeline
from .scheduled_instance import (
    ScheduledInstance,
    ScheduledActivityInstance,
    ScheduledDecisionInstance,
)
from .strength import Strength
from .study_amendment import StudyAmendment
from .study_amendment_reason import StudyAmendmentReason
from .study_amendment_impact import StudyAmendmentImpact
from .study_change import StudyChange
from .study_arm import StudyArm
from .study_cell import StudyCell
from .study_design import (
    InterventionalStudyDesign,
    ObservationalStudyDesign,
    StudyDesign,
)
from .study_element import StudyElement
from .study_epoch import StudyEpoch
from .identifier import (
    StudyIdentifier,
    AdministrableProductIdentifier,
    ReferenceIdentifier,
    MedicalDeviceIdentifier,
)
from .study_intervention import StudyIntervention
from .study_definition_document_version import StudyDefinitionDocumentVersion
from .study_definition_document import StudyDefinitionDocument
from .study_site import StudySite
from .study_title import StudyTitle
from .study_version import StudyVersion
from .study_role import StudyRole
from .study import Study
from .syntax_template import SyntaxTemplate
from .syntax_template_dictionary import SyntaxTemplateDictionary, ParameterMap
from .subject_enrollment import SubjectEnrollment
from .timing import Timing
from .transition_rule import TransitionRule
from .wrapper import Wrapper

__all__ = [
    "Abbreviation",
    "Activity",
    "Address",
    "AdministrableProduct",
    "AdministrableProductProperty",
    "Duration",
    "Administration",
    "AliasCode",
    "AnalysisPopulation",
    "AssignedPerson",
    "BiomedicalConceptCategory",
    "BiomedicalConceptProperty",
    "BiomedicalConceptSurrogate",
    "BiomedicalConcept",
    "Characteristic",
    "Code",
    "Condition",
    "DocumentContentReference",
    "EligibilityCriterion",
    "EligibilityCriterionItem",
    "Encounter",
    "Endpoint",
    "Estimand",
    "ExtensionAttribute",
    "ExtensionClass",
    "GeographicScope",
    "GovernanceDate",
    "Indication",
    "Ingredient",
    "IntercurrentEvent",
    "InterventionalStudyDesign",
    "Masking",
    "NarrativeContent",
    "NarrativeContentItem",
    "Objective",
    "ObservationalStudyDesign",
    "Organization",
    "Procedure",
    "Quantity",
    "Range",
    "ReferenceIdentifier",
    "ResponseCode",
    "ScheduleTimelineExit",
    "ScheduleTimeline",
    "ScheduledInstance",
    "ScheduledActivityInstance",
    "ScheduledDecisionInstance",
    "Strength",
    "StudyAmendment",
    "StudyAmendmentImpact",
    "StudyAmendmentReason",
    "StudyArm",
    "StudyCell",
    "StudyChange",
    "StudyCohort",
    "StudyDesignPopulation",
    "StudyDesign",
    "StudyElement",
    "StudyEpoch",
    "StudyIdentifier",
    "AdministrableProductIdentifier",
    "ReferenceIdentifier",
    "MedicalDeviceIdentifier",
    "StudyDefinitionDocumentVersion",
    "StudyDefinitionDocument",
    "StudyIntervention",
    "StudyRole",
    "StudySite",
    "StudyTitle",
    "StudyVersion",
    "Study",
    "SubjectEnrollment",
    "SyntaxTemplate",
    "SyntaxTemplateDictionary",
    "ParameterMap",
    "Timing",
    "TransitionRule",
    "Wrapper",
]



from usdm3.rules.library.rule_ddf00106 import RuleDDF00106 as V3Rule


class RuleDDF00106(V3Rule):
    def validate(self, config: dict) -> bool:
        return self._validate(
            config, ["InterventionalStudyDesign", "ObservationalStudyDesign"]
        )



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00053(RuleTemplate):
    """
    DDF00053: Within an encounter there must be no duplicate environmental settings.

    Applies to: Encounter
    Attributes: environmentalSetting
    """

    def __init__(self):
        super().__init__(
            "DDF00053",
            RuleTemplate.ERROR,
            "Within an encounter there must be no duplicate environmental settings.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.builder.builder import Builder
from usdm4.assembler.identification_assembler import IdentificationAssembler
from usdm4.assembler.population_assembler import PopulationAssembler
from usdm4.assembler.document_assembler import DocumentAssembler
from usdm4.assembler.study_design_assembler import StudyDesignAssembler
from usdm4.assembler.amendments_assembler import AmendmentsAssembler
from usdm4.assembler.study_assembler import StudyAssembler
from usdm4.assembler.timeline_assembler import TimelineAssembler
from usdm4.api.study import Study
from usdm4.api.wrapper import Wrapper
from usdm4.__info__ import __model_version__ as usdm_version


class Assembler:
    """
    Main assembler class responsible for orchestrating the assembly of a complete Study object
    from structured input data.

    The Assembler coordinates multiple specialized assemblers to process different domains
    of study data in the correct sequence, ensuring proper cross-references and dependencies
    are maintained throughout the assembly process.

    The assembly process follows a specific order to handle data dependencies:
    1. Identification (study IDs and versions)
    2. Documents (protocols and amendments)
    3. Populations (subject populations and analysis sets)
    4. Study Design (arms, epochs, activities, timelines)
    5. Study (core study information and final assembly)
    """

    MODULE = "usdm4.assembler.assembler.Assembler"

    def __init__(self, root_path: str, errors: Errors):
        self._errors = errors
        self._builder = Builder(root_path, self._errors)
        self._identification_assembler = IdentificationAssembler(
            self._builder, self._errors
        )
        self._population_assembler = PopulationAssembler(self._builder, self._errors)
        self._amendments_assembler = AmendmentsAssembler(self._builder, self._errors)
        self._document_assembler = DocumentAssembler(self._builder, self._errors)
        self._study_design_assembler = StudyDesignAssembler(self._builder, self._errors)
        self._study_assembler = StudyAssembler(self._builder, self._errors)
        self._timeline_assembler = TimelineAssembler(self._builder, self._errors)

    def clear(self):
        self._errors.clear()
        self._builder.clear()
        self._identification_assembler.clear()
        self._population_assembler.clear()
        self._amendments_assembler.clear()
        self._document_assembler.clear()
        self._study_design_assembler.clear()
        self._study_assembler.clear()
        self._timeline_assembler.clear()

    def execute(self, data: dict) -> None:
        """
        Executes the assembly process to build a complete Study object from structured data.

        Args:
            data (dict): A dictionary containing all the structured data needed to assemble a study.
                        The data parameter must have the following top-level structure:

                        {
                            "identification": dict,    # Study identification data (IDs, versions, etc.)
                            "document": dict,          # Document-related data (protocols, amendments, etc.)
                            "population": dict,        # Population definitions and analysis populations
                            "study_design": dict,      # Study design elements (arms, epochs, activities, etc.)
                            "study": dict             # Core study information (title, objectives, etc.)
                        }

                        Each top-level key corresponds to a specific domain of study data:

                        - "identification": Contains study identifiers, version information, and
                          regulatory identifiers needed to uniquely identify the study

                        - "document": Contains study definition documents, protocol versions,
                          amendments, and other document-related metadata

                        - "population": Contains population definitions, analysis populations,
                          eligibility criteria, and subject enrollment information

                        - "study_design": Contains the structural elements of the study design
                          including study arms, epochs, elements, activities, encounters,
                          procedures, and timeline information

                        - "study": Contains the core study metadata including titles, objectives,
                          indications, interventions, and high-level study characteristics

        Returns:
            None

        Note:
            The assembly process is sequential and interdependent:
            1. Identification data is processed first to establish study identity
            2. Document data is processed to set up protocol and amendment structure
            3. Population data is processed to define subject populations
            4. Study design data is processed (requires population data for references)
            5. Study data is processed last (requires all other components for assembly)
        """
        try:
            # Process identification data - establishes study identity and versioning
            self._identification_assembler.execute(data["identification"])

            # Process document data - sets up protocol documents and amendments
            self._document_assembler.execute(data["document"])

            # Process population data - defines subject populations and analysis sets
            self._population_assembler.execute(data["population"])

            # Process amendments data
            self._amendments_assembler.execute(data["amendments"])

            # Timelines data
            if "soa" in data:
                self._timeline_assembler.execute(data["soa"])

            # Process study design data - requires population assembler for cross-references
            self._study_design_assembler.execute(
                data["study_design"],
                self._population_assembler,
                self._timeline_assembler,
            )

            # Process core study data - requires all other assemblers for final assembly
            self._study_assembler.execute(
                data["study"],
                self._identification_assembler,
                self._study_design_assembler,
                self._document_assembler,
                self._population_assembler,
                self._amendments_assembler,
                self._timeline_assembler,
            )
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during assembler", e, location)

    @property
    def study(self) -> Study:
        return self._study_assembler.study

    def wrapper(self, name: str, version: str) -> Wrapper | None:
        try:
            params = {
                "study": self._study_assembler.study,
                "usdmVersion": usdm_version,
                "systemVersion": version,
                "systemName": name,
            }
            return self._builder.create(Wrapper, params)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during assembler", e, location)
            return None



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .study_amendment_reason import StudyAmendmentReason
from .study_change import StudyChange
from .study_amendment_impact import StudyAmendmentImpact
from .geographic_scope import GeographicScope
from .subject_enrollment import SubjectEnrollment
from .governance_date import GovernanceDate
from .comment_annotation import CommentAnnotation


class StudyAmendment(ApiBaseModelWithIdNameLabelAndDesc):
    number: str
    summary: str
    primaryReason: StudyAmendmentReason
    secondaryReasons: List[StudyAmendmentReason] = []
    changes: List[StudyChange] = []  # Not in API
    impacts: List[StudyAmendmentImpact] = []
    geographicScopes: List[GeographicScope]
    enrollments: List[SubjectEnrollment] = []
    dateValues: List[GovernanceDate] = []
    previousId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyAmendment"]



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code


class ProductOrganizationRole(ApiBaseModelWithIdNameLabelAndDesc):
    code: Code
    appliesToIds: List[str] = []
    organizationId: str
    instanceType: Literal["ProductOrganizationRole"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00017(RuleTemplate):
    """
    DDF00017: Within subject enrollment, the quantity must be a number or a percentage (i.e. the unit must be empty or %).

    Applies to: SubjectEnrollment
    Attributes: quantity
    """

    def __init__(self):
        super().__init__(
            "DDF00017",
            RuleTemplate.ERROR,
            "Within subject enrollment, the quantity must be a number or a percentage (i.e. the unit must be empty or %).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation


class StudyArm(ApiBaseModelWithIdNameLabelAndDesc):
    type: Code
    dataOriginDescription: str
    dataOriginType: Code
    populationIds: List[str] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyArm"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00062(RuleTemplate):
    """
    DDF00062: When specified, the upper limit of a timing window must be a non-negative duration in ISO 8601 format.

    Applies to: Timing
    Attributes: windowUpper
    """

    def __init__(self):
        super().__init__(
            "DDF00062",
            RuleTemplate.ERROR,
            "When specified, the upper limit of a timing window must be a non-negative duration in ISO 8601 format.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from .api_base_model import ApiBaseModelWithId


class DocumentContentReference(ApiBaseModelWithId):
    sectionNumber: str
    sectionTitle: str
    appliesToId: str
    instanceType: Literal["DocumentContentReference"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0191(RuleTemplate):
    """
    CHK0191: All abbreviations defined for a study version must be unique.

    Applies to: Abbreviation
    Attributes: abbreviation
    """

    def __init__(self):
        super().__init__(
            "CHK0191",
            RuleTemplate.ERROR,
            "All abbreviations defined for a study version must be unique.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .person_name import PersonName


class AssignedPerson(ApiBaseModelWithIdNameLabelAndDesc):
    personName: PersonName
    jobTitle: str
    organizationId: Union[str, None] = None
    instanceType: Literal["AssignedPerson"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00077(RuleTemplate):
    """
    DDF00077: If geographic scope type is global then no codes are expected to specify the specific area within scope while if it is not global then at least one code is expected to specify the specific area within scope.

    Applies to: GeographicScope
    Attributes: code
    """

    def __init__(self):
        super().__init__(
            "DDF00077",
            RuleTemplate.ERROR,
            "If geographic scope type is global then no codes are expected to specify the specific area within scope while if it is not global then at least one code is expected to specify the specific area within scope.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00100(RuleTemplate):
    """
    DDF00100: Within a study version, there must be no more than one title of each type.

    Applies to: StudyVersion
    Attributes: titles
    """

    def __init__(self):
        super().__init__(
            "DDF00100",
            RuleTemplate.ERROR,
            "Within a study version, there must be no more than one title of each type.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdAndName


class NarrativeContentItem(ApiBaseModelWithIdAndName):
    text: str
    instanceType: Literal["NarrativeContentItem"]


class NarrativeContent(ApiBaseModelWithIdAndName):
    sectionNumber: Union[str, None] = None
    sectionTitle: Union[str, None] = None
    displaySectionNumber: bool
    displaySectionTitle: bool
    childIds: List[str] = []
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    contentItemId: Union[str, None] = None
    instanceType: Literal["NarrativeContent"]



from typing import List, Literal, Union
from .activity import Activity
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode
from .biospecimen_retention import BiospecimenRetention
from .code import Code
from .encounter import Encounter
from .study_cell import StudyCell
from .indication import Indication
from .study_arm import StudyArm
from .study_epoch import StudyEpoch
from .study_element import StudyElement
from .population_definition import StudyDesignPopulation
from .eligibility_criterion import EligibilityCriterion
from .analysis_population import AnalysisPopulation
from .objective import Objective
from .schedule_timeline import ScheduleTimeline
from .estimand import Estimand
from .comment_annotation import CommentAnnotation


class StudyDesign(ApiBaseModelWithIdNameLabelAndDesc):
    studyType: Union[Code, None] = None
    studyPhase: Union[AliasCode, None] = None
    therapeuticAreas: List[Code] = []
    characteristics: List[Code] = []
    encounters: List[Encounter] = []
    activities: List[Activity] = []
    arms: List[StudyArm]
    studyCells: List[StudyCell]
    rationale: str
    epochs: List[StudyEpoch]
    elements: List[StudyElement] = []
    estimands: List[Estimand] = []
    indications: List[Indication] = []
    studyInterventionIds: List[str] = []
    objectives: List[Objective] = []
    population: StudyDesignPopulation
    scheduleTimelines: List[ScheduleTimeline] = []
    biospecimenRetentions: List[BiospecimenRetention] = []
    documentVersionIds: List[str] = []
    eligibilityCriteria: List[EligibilityCriterion] = []
    analysisPopulations: List[AnalysisPopulation] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyDesign"]

    def main_timeline(self):
        return next(
            (item for item in self.scheduleTimelines if item.mainTimeline), None
        )

    def phase(self) -> Code:
        try:
            return self.studyPhase.standardCode
        except Exception:
            return None

    def phase_as_text(self) -> str:
        code = self.phase()
        return code.decode if code else ""

    # def soa(self, timeline_name: str) -> list:
    #     timeline = next(
    #         (x for x in self.scheduleTimelines if x.name == timeline_name), None
    #     )
    #     return timeline.soa() if timeline else None

    # def main_soa(self) -> list:
    #     timeline = next((x for x in self.scheduleTimelines if x.mainTimeline), None)
    #     return timeline.soa() if timeline else None

    def first_activity(self) -> Activity:
        return next((x for x in self.activities if not x.previousId and x.nextId), None)

    def find_activity(self, id: str) -> Activity:
        return next((x for x in self.activities if x.id == id), None)

    def activity_list(self) -> list:
        items = []
        item = self.first_activity()
        while item:
            items.append(item)
            item = self.find_activity(item.nextId)
        return items

    def find_epoch(self, id: str) -> StudyEpoch:
        return next((x for x in self.epochs if x.id == id), None)

    def find_encounter(self, id: str) -> Encounter:
        return next((x for x in self.encounters if x.id == id), None)

    def find_timeline(self, id: str) -> ScheduleTimeline:
        return next((x for x in self.scheduleTimelines if x.id == id), None)

    def find_analysis_population(self, id: str) -> AnalysisPopulation:
        return next((x for x in self.analysisPopulations if x.id == id), None)

    def criterion_map(self) -> dict[EligibilityCriterion]:
        return {x.id: x for x in self.eligibilityCriteria}


class InterventionalStudyDesign(StudyDesign):
    subTypes: List[Code] = []
    model: Code
    intentTypes: List[Code] = []
    blindingSchema: Union[AliasCode, None] = None
    instanceType: Literal["InterventionalStudyDesign"]


class ObservationalStudyDesign(StudyDesign):
    subTypes: List[Code] = []
    model: Code
    timePerspective: Code
    samplingMethod: Union[Code, None] = None
    instanceType: Literal["ObservationalStudyDesign"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00047(RuleTemplate):
    """
    DDF00047: A study cell must only reference elements that are defined within the same study design as the study cell.

    Applies to: StudyCell
    Attributes: elements
    """

    def __init__(self):
        super().__init__(
            "DDF00047",
            RuleTemplate.ERROR,
            "A study cell must only reference elements that are defined within the same study design as the study cell.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00126(RuleTemplate):
    """
    DDF00126: Cardinalities must be as defined in the USDM schema based on the API specification (i.e., required properties have at least one value and single-value properties are not lists).

    Applies to: All
    Attributes: All
    """

    def __init__(self):
        super().__init__(
            "DDF00126",
            RuleTemplate.ERROR,
            "Cardinalities must be as defined in the USDM schema based on the API specification (i.e., required properties have at least one value and single-value properties are not lists).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0248(RuleTemplate):
    """
    CHK0248: Narrative content must only reference narrative content that is specified within the same study definition document version.

    Applies to: NarrativeContent
    Attributes: next, previous, children
    """

    def __init__(self):
        super().__init__(
            "CHK0248",
            RuleTemplate.ERROR,
            "Narrative content must only reference narrative content that is specified within the same study definition document version.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from pydantic import Field
from .api_base_model import ApiBaseModelWithId
from .comment_annotation import CommentAnnotation


class Abbreviation(ApiBaseModelWithId):
    abbreviatedText: str = Field(min_length=1)
    expandedText: str = Field(min_length=1)
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Abbreviation"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00122(RuleTemplate):
    """
    DDF00122: A study design's characteristics must be specified according to the Study design characteristics (C207416) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: StudyDesign
    Attributes: characteristics
    """

    def __init__(self):
        super().__init__(
            "DDF00122",
            RuleTemplate.ERROR,
            "A study design's characteristics must be specified according to the Study design characteristics (C207416) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00115(RuleTemplate):
    """
    DDF00115: Every study version must have a title of type \"Official Study Title\".

    Applies to: StudyVersion
    Attributes: titles
    """

    def __init__(self):
        super().__init__(
            "DDF00115",
            RuleTemplate.ERROR,
            'Every study version must have a title of type "Official Study Title".',
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, Literal
from .api_base_model import ApiBaseModelWithIdAndName
from .quantity_range import Quantity
from .code import Code


class AdministrableProductProperty(ApiBaseModelWithIdAndName):
    text: str
    type: Code
    quantity: Union[Quantity, None] = None
    instanceType: Literal["AdministrableProductProperty"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00019(RuleTemplate):
    """
    DDF00019: A scheduled activity/decision instance must not refer to itself as its default condition.

    Applies to: ScheduledActivityInstance, ScheduledDecisionInstance
    Attributes: defaultCondition
    """

    def __init__(self):
        super().__init__(
            "DDF00019",
            RuleTemplate.ERROR,
            "A scheduled activity/decision instance must not refer to itself as its default condition.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation


class Indication(ApiBaseModelWithIdNameLabelAndDesc):
    codes: List[Code] = []
    isRareDisease: bool
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Indication"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00010(RuleTemplate):
    """
    DDF00010: The names of all child instances of the same parent class must be unique.

    Applies to: All
    Attributes: name
    """

    def __init__(self):
        super().__init__(
            "DDF00010",
            RuleTemplate.ERROR,
            "The names of all child instances of the same parent class must be unique.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00020(RuleTemplate):
    """
    DDF00020: If the reason for a study amendment is 'Other' then this must be specified (attribute reasonOther must be completed)

    Applies to: StudyAmendmentReason
    Attributes: code, otherReason
    """

    def __init__(self):
        super().__init__(
            "DDF00020",
            RuleTemplate.ERROR,
            "If the reason for a study amendment is 'Other' then this must be specified (attribute reasonOther must be completed)",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent
from .comment_annotation import CommentAnnotation


class StudyDefinitionDocumentVersion(ApiBaseModelWithId):
    version: str
    status: Code
    dateValues: List[GovernanceDate] = []
    contents: List[NarrativeContent] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyDefinitionDocumentVersion"]

    def narrative_content_in_order(self):
        sections = []
        narrative_content = self._first_narrative_content()
        while narrative_content:
            sections.append(narrative_content)
            narrative_content = next(
                (x for x in self.contents if x.id == narrative_content.nextId), None
            )
        return sections

    def find_narrative_content(self, id: str) -> NarrativeContent:
        try:
            map = self.narraitve_content_map
            return map[id]
        except Exception:
            return None

    def narraitve_content_map(self) -> dict:
        return {x.id: x for x in self.contents}

    def _first_narrative_content(self) -> NarrativeContent:
        return next((x for x in self.contents if not x.previousId and x.nextId), None)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00007(RuleTemplate):
    """
    DDF00007: If timing type is \"Fixed Reference\" then it must point to only one scheduled instance (e.g. attribute relativeToScheduledInstance must be equal to relativeFromScheduledInstance or it must be missing).

    Applies to: Timing
    Attributes: relativeToScheduledInstance
    """

    def __init__(self):
        super().__init__(
            "DDF00007",
            RuleTemplate.ERROR,
            'If timing type is "Fixed Reference" then it must point to only one scheduled instance (e.g. attribute relativeToScheduledInstance must be equal to relativeFromScheduledInstance or it must be missing).',
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00109(RuleTemplate):
    """
    DDF00109: A study element must only reference study interventions that are defined within the same study design as the study element.

    Applies to: StudyElement
    Attributes: studyInterventions
    """

    def __init__(self):
        super().__init__(
            "DDF00109",
            RuleTemplate.ERROR,
            "A study element must only reference study interventions that are defined within the same study design as the study element.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import json
from typing import Union, List, Literal
from pydantic import BaseModel, Field
from .serialize import serialize_as_json


class Extension(BaseModel):
    id: str = Field(min_length=1)
    url: str

    def to_json(self):
        return json.dumps(self, default=serialize_as_json)

class BaseDataType(BaseModel):
    id: str = Field(min_length=1)

    def to_json(self):
        return json.dumps(self, default=serialize_as_json)

class BaseCode(BaseDataType):
    code: str
    codeSystem: str
    codeSystemVersion: str
    decode: str
    instanceType: Literal["Code"]
    extensionAttributes: List["ExtensionAttribute"] = []


class BaseAliasCode(BaseDataType):
    standardCode: BaseCode
    standardCodeAliases: List[BaseCode] = []
    instanceType: Literal["AliasCode"]
    extensionAttributes: List["ExtensionAttribute"] = []


class BaseQuantity(BaseDataType):
    value: float
    unit: Union[BaseAliasCode, None] = None
    instanceType: Literal["Quantity"]
    extensionAttributes: List["ExtensionAttribute"] = []


class BaseRange(BaseDataType):
    minValue: BaseQuantity
    maxValue: BaseQuantity
    isApproximate: bool
    instanceType: Literal["Range"]
    extensionAttributes: List["ExtensionAttribute"] = []


class ExtensionAttribute(Extension):
    # values or extension attributes, never both.
    valueString: Union[str, None] = None
    valueBoolean: Union[bool, None] = None
    valueInteger: Union[int, None] = None
    valueId: Union[str, None] = None
    valueQuantity: Union[BaseQuantity, None] = None
    valueRange: Union[BaseRange, None] = None
    valueCode: Union[BaseCode, None] = None
    valueAliasCode: Union[BaseAliasCode, None] = None
    valueExtensionClass: Union["ExtensionClass", None] = None
    extensionAttributes: List["ExtensionAttribute"] = []
    instanceType: Literal["ExtensionAttribute"]


class ExtensionClass(Extension):
    extensionAttributes: List["ExtensionAttribute"] = []
    instanceType: Literal["ExtensionClass"]



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.assembler.population_assembler import PopulationAssembler
from usdm4.assembler.timeline_assembler import TimelineAssembler
from usdm4.assembler.encoder import Encoder
from usdm4.builder.builder import Builder
from usdm4.api.study_design import InterventionalStudyDesign


class StudyDesignAssembler(BaseAssembler):
    """
    Assembler responsible for creating InterventionalStudyDesign objects from study design data.

    This assembler processes study design information including intervention models, study phases,
    arms, epochs, cells, and other structural elements that define how the study is conducted.
    It creates the core study design structure that serves as the framework for the clinical trial.
    """

    MODULE = "usdm4.assembler.study_design_assembler.StudyDesignAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the StudyDesignAssembler.

        Args:
            builder (Builder): The builder instance for creating USDM objects
            errors (Errors): Error handling instance for logging issues
        """
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        self._study_design = None

    def execute(
        self,
        data: dict,
        population_assembler: PopulationAssembler,
        timeline_assembler: TimelineAssembler,
    ) -> None:
        """
        Creates an InterventionalStudyDesign object from study design data.

        Args:
            data (dict): A dictionary containing study design information.
                        The data parameter must have the following structure:

                        {
                            "label": str,              # Human-readable label for the study design
                            "rationale": str,          # Rationale or justification for this design
                            "trial_phase": str,        # Clinical trial phase (e.g., "Phase I", "Phase II")
                            # Additional optional fields may include:
                            # "description": str,       # Detailed description of the study design
                            # "intervention_model": str, # Type of intervention model (parallel, crossover, etc.)
                            # "arms": list,             # List of study arms/treatment groups
                            # "epochs": list,           # List of study epochs/periods
                            # "cells": list,            # List of study cells (arm-epoch combinations)
                            # "objectives": list,       # List of study objectives
                            # "estimands": list,        # List of estimands for analysis
                            # "interventions": list,    # List of study interventions
                            # "analysis_populations": list,  # List of analysis population definitions
                        }

                        Required fields:
                        - "label": Display name for the study design
                        - "rationale": Explanation for why this design was chosen
                        - "trial_phase": The clinical development phase of the study

            population_assembler (PopulationAssembler): Assembler containing the study population
                definition that will be referenced by this study design
            timeline_assembler (TimelineAssembler): Assembler containing the timelines

        Returns:
            None: The created study design is stored in self._study_design property

        Note:
            The current implementation creates a basic interventional study design with:
            - Default parallel study intervention model (CDISC code C82639)
            - Empty lists for arms, cells, epochs, objectives, estimands, interventions, and analysis populations
            - Reference to the population from the population_assembler
            - Study phase from the input data
        """
        try:
            # Get CDISC code for parallel study intervention model (default)
            intervention_model_code = self._builder.cdisc_code(
                "C82639", "Parallel Study"
            )

            # Create the InterventionalStudyDesign object with basic structure
            self._study_design = self._builder.create(
                InterventionalStudyDesign,
                {
                    "name": self._label_to_name(data["label"]),
                    "label": data["label"],
                    "description": "A study design",
                    "rationale": data["rationale"],
                    "model": intervention_model_code,  # Default to parallel study model
                    "arms": [],  # Empty arms list (future enhancement)
                    "studyCells": [],  # Empty cells list (future enhancement)
                    "epochs": timeline_assembler.epochs,
                    "encounters": timeline_assembler.encounters,
                    "activities": timeline_assembler.activities,
                    "population": population_assembler.population,
                    "objectives": [],  # Empty objectives list (future enhancement)
                    "estimands": [],  # Empty estimands list (future enhancement)
                    "studyInterventions": [],  # Empty interventions list (future enhancement)
                    "analysisPopulations": [],  # Empty analysis populations list (future enhancement)
                    "studyPhase": self._encoder.phase(data["trial_phase"]),
                    "scheduleTimelines": timeline_assembler.timelines,
                },
            )
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception(
                "Failed during creation of study design", e, location
            )

    @property
    def study_design(self) -> InterventionalStudyDesign:
        return self._study_design



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00124(RuleTemplate):
    """
    DDF00124: Referenced items in a parameter map must be available elsewhere in the data model.

    Applies to: ParameterMap
    Attributes: reference
    """

    def __init__(self):
        super().__init__(
            "DDF00124",
            RuleTemplate.ERROR,
            "Referenced items in a parameter map must be available elsewhere in the data model.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0202(RuleTemplate):
    """
    CHK0202: If a dose is specified then a corresponding frequency must also be specified.

    Applies to: Administration
    Attributes: dose
    """

    def __init__(self):
        super().__init__(
            "CHK0202",
            RuleTemplate.ERROR,
            "If a dose is specified then a corresponding frequency must also be specified.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union
from .api_base_model import ApiBaseModel
from .study import Study


class Wrapper(ApiBaseModel):
    study: Study
    usdmVersion: str
    systemName: Union[str, None] = None
    systemVersion: Union[str, None] = None



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00023(RuleTemplate):
    """
    DDF00023: To ensure consistent ordering, when both previous and next attributes are available within an entity the previous id value must match the next id value of the referred instance.

    Applies to: Activity, EligibilityCriterion, Encounter, NarrativeContent, StudyEpoch
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00023",
            RuleTemplate.ERROR,
            "To ensure consistent ordering, when both previous and next attributes are available within an entity the previous id value must match the next id value of the referred instance.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00059(RuleTemplate):
    """
    DDF00059: Within a study intervention, if more intervention codes are defined, they must be distinct.

    Applies to: StudyIntervention
    Attributes: codes
    """

    def __init__(self):
        super().__init__(
            "DDF00059",
            RuleTemplate.ERROR,
            "Within a study intervention, if more intervention codes are defined, they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00003(RuleTemplate):
    """
    DDF00003: If the duration of an administration will vary, a quantity is not expected for the administration duration and vice versa.

    Applies to: AdministrationDuration
    Attributes: quantity
    """

    def __init__(self):
        super().__init__(
            "DDF00003",
            RuleTemplate.ERROR,
            "If the duration of an administration will vary, a quantity is not expected for the administration duration and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0061(RuleTemplate):
    """
    CHK0061: Each specified biomedical concept category is expected to be referenced by an activity.

    Applies to: Activity
    Attributes: bcCategories
    """

    def __init__(self):
        super().__init__(
            "CHK0061",
            RuleTemplate.ERROR,
            "Each specified biomedical concept category is expected to be referenced by an activity.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .biomedical_concept_property import BiomedicalConceptProperty
from .comment_annotation import CommentAnnotation


class BiomedicalConcept(ApiBaseModelWithIdNameAndLabel):
    synonyms: List[str] = []
    reference: str
    properties: List[BiomedicalConceptProperty] = []
    code: AliasCode
    notes: List[CommentAnnotation] = []
    instanceType: Literal["BiomedicalConcept"]



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .intercurrent_event import IntercurrentEvent
from .comment_annotation import CommentAnnotation


class Estimand(ApiBaseModelWithIdNameLabelAndDesc):
    populationSummary: str
    analysisPopulationId: str
    interventionIds: List[str]
    variableOfInterestId: str
    intercurrentEvents: List[IntercurrentEvent]
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Estimand"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0250(RuleTemplate):
    """
    CHK0250: Sourcing must not be defined for an administrable product which is only referenced as an embedded product for a medical device.

    Applies to: AdminstrableProduct
    Attributes: sourcing
    """

    def __init__(self):
        super().__init__(
            "CHK0250",
            RuleTemplate.ERROR,
            "Sourcing must not be defined for an administrable product which is only referenced as an embedded product for a medical device. ",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00075(RuleTemplate):
    """
    DDF00075: An activity is expected to refer to at least one procedure, biomedical concept, biomedical concept category or biomedical concept surrogate.

    Applies to: Activity
    Attributes: definedProcedures, biomedicalConcepts, bcCategories, bcSurrogates
    """

    def __init__(self):
        super().__init__(
            "DDF00075",
            RuleTemplate.ERROR,
            "An activity is expected to refer to at least one procedure, biomedical concept, biomedical concept category or biomedical concept surrogate.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00078(RuleTemplate):
    """
    DDF00078: If a transition start rule is defined then an end rule is expected and vice versa.

    Applies to: StudyElement, Encounter
    Attributes: transitionStartRule, transitionEndRule
    """

    def __init__(self):
        super().__init__(
            "DDF00078",
            RuleTemplate.ERROR,
            "If a transition start rule is defined then an end rule is expected and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0272(RuleTemplate):
    """
    CHK0272: An observational study (including patient registries) must be specified using the ObservationalStudyDesign class.

    Applies to: ObservationalStudyDesign
    Attributes: studyType
    """

    def __init__(self):
        super().__init__(
            "CHK0272",
            RuleTemplate.ERROR,
            "An observational study (including patient registries) must be specified using the ObservationalStudyDesign class.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00073(RuleTemplate):
    """
    DDF00073: Only one version of any code system is expected to be used within a study version.

    Applies to: Code
    Attributes: codeSystemVersion
    """

    def __init__(self):
        super().__init__(
            "DDF00073",
            RuleTemplate.ERROR,
            "Only one version of any code system is expected to be used within a study version.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00018(RuleTemplate):
    """
    DDF00018: An instance of a class must not reference itself as one of its own children.

    Applies to: BiomedicalConceptCategory, StudyProtocolDocumentVersion, StudyDefinitionDocumentVersion, NarrativeContent, Activity
    Attributes: children
    """

    def __init__(self):
        super().__init__(
            "DDF00018",
            RuleTemplate.ERROR,
            "An instance of a class must not reference itself as one of its own children.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00014(RuleTemplate):
    """
    DDF00014: A biomedical concept category is expected to have at least a member or a child.

    Applies to: BiomedicalConceptCategory
    Attributes: members, children
    """

    def __init__(self):
        super().__init__(
            "DDF00014",
            RuleTemplate.ERROR,
            "A biomedical concept category is expected to have at least a member or a child.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation


class BiomedicalConceptCategory(ApiBaseModelWithIdNameLabelAndDesc):
    childIds: List[str] = []
    memberIds: List[str] = []
    code: Union[AliasCode, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["BiomedicalConceptCategory"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0231(RuleTemplate):
    """
    CHK0231: Narrative content item text is expected to be HTML formatted.

    Applies to: NarrativeContentItem
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "CHK0231",
            RuleTemplate.ERROR,
            "Narrative content item text is expected to be HTML formatted.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code


class Timing(ApiBaseModelWithIdNameLabelAndDesc):
    type: Code
    value: str
    valueLabel: str
    relativeToFrom: Code
    relativeFromScheduledInstanceId: str
    relativeToScheduledInstanceId: Union[str, None] = None
    windowLower: Union[str, None] = None
    windowUpper: Union[str, None] = None
    windowLabel: Union[str, None] = None
    instanceType: Literal["Timing"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation
from .transition_rule import TransitionRule


class Encounter(ApiBaseModelWithIdNameLabelAndDesc):
    type: Code
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    scheduledAtId: Union[str, None] = None
    environmentalSettings: List[Code] = []
    contactModes: List[Code] = []
    transitionStartRule: Union[TransitionRule, None] = None
    transitionEndRule: Union[TransitionRule, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Encounter"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00084(RuleTemplate):
    """
    DDF00084: Within a study design there must be exactly one objective with level 'Primary Objective'.

    Applies to: Objective
    Attributes: level
    """

    def __init__(self):
        super().__init__(
            "DDF00084",
            RuleTemplate.ERROR,
            "Within a study design there must be exactly one objective with level 'Primary Objective'.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0181(RuleTemplate):
    """
    CHK0181: When included in text, references to items stored elsewhere in the data model must be specified in the correct format. They must start with '<usdm:ref', end with either '/>' or '></usdm:ref>', and must contain 'klass=\"KlassName\"',  'id=\"idValue\"', and 'attribute=\"attributeName\"/>' in any order (where \"ClassName\" and \"attributeName\" contain only letters in upper or lower case).

    Applies to: NarrativeContentItem
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "CHK0181",
            RuleTemplate.ERROR,
            "When included in text, references to items stored elsewhere in the data model must be specified in the correct format. They must start with '<usdm:ref', end with either '/>' or '></usdm:ref>', and must contain 'klass=\"KlassName\"',  'id=\"idValue\"', and 'attribute=\"attributeName\"/>' in any order (where \"ClassName\" and \"attributeName\" contain only letters in upper or lower case).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00099(RuleTemplate):
    """
    DDF00099: All epochs are expected to be referred to from a scheduled Activity Instance.

    Applies to: ScheduledActivityInstance
    Attributes: epoch
    """

    def __init__(self):
        super().__init__(
            "DDF00099",
            RuleTemplate.ERROR,
            "All epochs are expected to be referred to from a scheduled Activity Instance.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import json
import os


class Library:
    BASE_PATH = "ct/iso/iso3166"

    def __init__(self, root_path: str):
        self.system = "ISO 3166 1 alpha3"
        self.version = "2020-08"
        self.filepath = os.path.join(root_path, self.BASE_PATH, "iso3166.json")
        self.db = None

    def load(self) -> None:
        f = open(self.filepath)
        self.db = json.load(f)

    def code_or_decode(self, text: str) -> tuple[str | None, str | None]:
        code, name = self._get_decode(text)
        if not code:
            code, name = self._get_code(text)
        return code, name

    def code(self, decode: str) -> tuple[str | None, str | None]:
        return self._get_code(decode)

    def decode(self, code: str) -> tuple[str | None, str | None]:
        return self._get_decode(code)

    def region_code(self, decode: str) -> tuple[str | None, str | None]:
        code, decode = self._get_region_decode(decode)
        return code, decode

    def _get_decode(self, code: str) -> tuple[str | None, str | None]:
        field = "alpha-2" if len(code) == 2 else "alpha-3"
        entry = next((item for item in self.db if item[field] == code), None)
        return (None, None) if entry is None else (entry["alpha-3"], entry["name"])

    def _get_code(self, decode: str) -> tuple[str | None, str | None]:
        entry = next(
            (item for item in self.db if item["name"].upper() == decode.upper()), None
        )
        return (None, None) if entry is None else (entry["alpha-3"], entry["name"])

    def _get_region_decode(self, decode: str) -> tuple[str | None, str | None]:
        for scope in ["region", "sub-region", "intermediate-region"]:
            entry = next(
                (item for item in self.db if item[scope].upper() == decode.upper()),
                None,
            )
            if entry:
                return entry[f"{scope}-code"], entry[scope]
        return None, None



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00103(RuleTemplate):
    """
    DDF00103: Within a document version, the specified section numbers for narrative content must be unique.

    Applies to: NarrativeContent
    Attributes: sectionNumber
    """

    def __init__(self):
        super().__init__(
            "DDF00103",
            RuleTemplate.ERROR,
            "Within a document version, the specified section numbers for narrative content must be unique.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, List, Literal, Annotated
from annotated_types import Len
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .quantity_range import Quantity, Range
from .characteristic import Characteristic
from .comment_annotation import CommentAnnotation


class PopulationDefinition(ApiBaseModelWithIdNameLabelAndDesc):
    includesHealthySubjects: bool
    plannedEnrollmentNumber: Union[Quantity, Range, None] = None
    plannedCompletionNumber: Union[Quantity, Range, None] = None
    plannedSex: Annotated[List[Code], Len(min_length=0, max_length=2)] = []
    criterionIds: List[str] = []
    plannedAge: Union[Range, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["PopulationDefinition"]


class StudyCohort(PopulationDefinition):
    characteristics: List[Characteristic] = []
    indicationIds: List[str] = []
    instanceType: Literal["StudyCohort"]


class StudyDesignPopulation(PopulationDefinition):
    cohorts: List[StudyCohort] = []
    instanceType: Literal["StudyDesignPopulation"]



from typing import Literal, List
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .comment_annotation import CommentAnnotation


class StudyAmendmentImpact(ApiBaseModelWithId):
    type: Code
    text: str
    isSubstantial: bool
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyAmendmentImpact"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00080(RuleTemplate):
    """
    DDF00080: All scheduled activity instances are expected to refer to an epoch.

    Applies to: ScheduledActivityInstance
    Attributes: epoch
    """

    def __init__(self):
        super().__init__(
            "DDF00080",
            RuleTemplate.ERROR,
            "All scheduled activity instances are expected to refer to an epoch.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import json
import pathlib
from typing_extensions import deprecated
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.rules.rules_validation import RulesValidation4
from usdm3.rules.rules_validation_results import RulesValidationResults
from usdm4.api.wrapper import Wrapper
from usdm4.convert.convert import Convert
from usdm4.builder.builder import Builder
from usdm4.assembler.assembler import Assembler


class USDM4:
    MODULE = "usdm4.USDM4"

    def __init__(self):
        self.root = self._root_path()
        self.validator = RulesValidation4(self.root)

    def validate(self, file_path: str) -> RulesValidationResults:
        return self.validator.validate(file_path)

    def convert(self, file_path: str) -> Wrapper:
        with open(file_path, "r") as file:
            data = json.load(file)
        return Convert.convert(data)

    def builder(self, errors: Errors) -> Builder:
        return Builder(self.root, errors)

    def assembler(self, errors: Errors) -> Assembler:
        return Assembler(self.root, errors)

    def minimum(
        self, study_name: str, sponsor_id: str, version: str, errors: Errors
    ) -> Wrapper:
        return Builder(self.root, errors).minimum(study_name, sponsor_id, version)

    @deprecated("Use the 'load' or the 'loadd' methods")
    def from_json(self, data: dict) -> Wrapper:
        return Wrapper.model_validate(data)

    def loadd(self, data: dict, errors: Errors) -> Wrapper | None:
        try:
            return Wrapper.model_validate(data)
        except Exception as e:    
            errors.exception("Failed to load a dict into USDM", e, KlassMethodLocation(self.MODULE, "from_dict"))
            return None

    def load(self, filepath: str, errors: Errors) -> Wrapper | None:
        try:
            data = None
            with open(filepath, "r") as f:
                data = json.load(f)
                f.close()
            return Wrapper.model_validate(data)
        except Exception as e:    
            errors.exception("Failed to load file '{filepath}' into USDM", e, KlassMethodLocation(self.MODULE, "load"))
            return None

    def _root_path(self) -> str:
        return pathlib.Path(__file__).parent.resolve()



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00114(RuleTemplate):
    """
    DDF00114: If specified, the context of a condition must point to a valid instance in the activity or scheduled activity instance class.

    Applies to: Condition
    Attributes: context
    """

    def __init__(self):
        super().__init__(
            "DDF00114",
            RuleTemplate.ERROR,
            "If specified, the context of a condition must point to a valid instance in the activity or scheduled activity instance class.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_ddfsdw001 import RuleDDFSDW001 as V3Rule
from usdm4.__info__ import __model_version__ as model_version


class RuleDDFSDW001(V3Rule):
    def __init__(self):
        super().__init__()
        self._rule_text = "The version in the wrapper should be set to 4.0.0"

    def validate(self, config: dict) -> bool:
        return self._validate_version(config, model_version)



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.assembler.encoder import Encoder
from usdm4.builder.builder import Builder
from usdm4.api.quantity_range import Quantity
from usdm4.api.geographic_scope import GeographicScope
from usdm4.api.subject_enrollment import SubjectEnrollment
from usdm4.api.study_amendment_reason import StudyAmendmentReason
from usdm4.api.study_amendment import StudyAmendment


class AmendmentsAssembler(BaseAssembler):
    MODULE = "usdm4.assembler.amendments_assembler.AmenementsAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        self._amendment = None

    def execute(self, data: dict) -> None:
        try:
            if data:
                self._amendment = self._create_amendment(data)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of amendments", e, location)

    @property
    def amendment(self) -> StudyAmendment:
        return self._amendment

    def _create_amendment(self, data: dict) -> StudyAmendment:
        try:
            # print(f"DATA: {data}")
            reason = {}
            global_code = self._builder.cdisc_code("C68846", "Global")
            global_scope = self._builder.create(GeographicScope, {"type": global_code})
            for k, item in data["reasons"].items():
                # print(f"REASON_CODE: {k} = {item}")
                reason[k] = self._builder.create(
                    StudyAmendmentReason, self._encoder.amendment_reason(item)
                )
            impact = data["impact"]["safety"] or data["impact"]["reliability"]
            # print(f"IMPACT: {impact}")
            params = {
                "name": "AMENDMENT 1",
                "number": "1",
                "summary": data["summary"],
                "substantialImpact": impact,
                "primaryReason": reason["primary"],
                "secondaryReasons": [reason["secondary"]],
                "enrollments": [self._create_enrollment(data)],
                "geographicScopes": [global_scope],
            }
            return self._builder.create(StudyAmendment, params)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of amendments", e, location)
            return None

    def _create_enrollment(self, data: dict):
        try:
            global_code = self._builder.cdisc_code("C68846", "Global")
            params = {
                "type": global_code,
                "code": None,
            }
            geo_scope = self._builder.create(GeographicScope, params)
            if "enrollment" in data:
                unit_alias = None
                if data["enrollment"]["unit"] == "%":
                    unit_code = self._builder.cdisc_code("C25613", "Percentage")
                    unit_alias = (
                        self._builder.alias_code(unit_code) if unit_code else None
                    )
                quantity = self._builder.create(
                    Quantity, {"value": data["enrollment"]["value"], "unit": unit_alias}
                )
                params = {
                    "name": "ENROLLMENT",
                    "forGeographicScope": geo_scope,
                    "quantity": quantity,
                }
            else:
                quantity = self._builder.create(Quantity, {"value": 0, "unit": None})
                params = {
                    "name": "ENROLLMENT",
                    "forGeographicScope": geo_scope,
                    "quantity": quantity,
                }
            return self._builder.create(SubjectEnrollment, params)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of enrollments", e, location)
            return None



from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from typing import Literal


class TransitionRule(ApiBaseModelWithIdNameLabelAndDesc):
    text: str
    instanceType: Literal["TransitionRule"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0221(RuleTemplate):
    """
    CHK0221: A unit must be specified for every strength denominator and numerator

    Applies to: Strength
    Attributes: numerator, denominator
    """

    def __init__(self):
        super().__init__(
            "CHK0221",
            RuleTemplate.ERROR,
            "A unit must be specified for every strength denominator and numerator",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00016(RuleTemplate):
    """
    DDF00016: A specified condition for assessments must apply to at least to a procedure, biomedical concept, biomedical concept surrogate, biomedical concept category or a whole activity.

    Applies to: Condition
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "DDF00016",
            RuleTemplate.ERROR,
            "A specified condition for assessments must apply to at least to a procedure, biomedical concept, biomedical concept surrogate, biomedical concept category or a whole activity.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00101(RuleTemplate):
    """
    DDF00101: Within a study design, if study type is Interventional then at least one intervention is expected to be referenced from a procedure.

    Applies to: Procedure
    Attributes: studyIntervention
    """

    def __init__(self):
        super().__init__(
            "DDF00101",
            RuleTemplate.ERROR,
            "Within a study design, if study type is Interventional then at least one intervention is expected to be referenced from a procedure.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00145(RuleTemplate):
    """
    DDF00145: A unit must be coded according to the extensible unit (C71620) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: Range, Quantity
    Attributes: unit
    """

    def __init__(self):
        super().__init__(
            "DDF00145",
            RuleTemplate.ERROR,
            "A unit must be coded according to the extensible unit (C71620) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from simple_error_log.errors import Errors
from usdm4.builder.builder import Builder


class BaseAssembler:
    MODULE = "usdm4.assembler.base_assembler.BaseAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        self._errors = errors
        self._builder = builder

    def execute(self, data: dict) -> None:
        pass

    def _label_to_name(self, text: str) -> str:
        return text.upper().replace(" ", "-")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00042(RuleTemplate):
    """
    DDF00042: The range specified for a planned age is not expected to be approximate.

    Applies to: Range
    Attributes: isApproximate
    """

    def __init__(self):
        super().__init__(
            "DDF00042",
            RuleTemplate.ERROR,
            "The range specified for a planned age is not expected to be approximate.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0255(RuleTemplate):
    """
    CHK0255: A product organization role is expected to apply to at least one medical device or administrable product.

    Applies to: ProductOrganizationRole
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "CHK0255",
            RuleTemplate.ERROR,
            "A product organization role is expected to apply to at least one medical device or administrable product.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0192(RuleTemplate):
    """
    CHK0192: The long names for all abbreviations defined for a study version are expected to be unique.

    Applies to: Abbreviation
    Attributes: longName
    """

    def __init__(self):
        super().__init__(
            "CHK0192",
            RuleTemplate.ERROR,
            "The long names for all abbreviations defined for a study version are expected to be unique.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00127(RuleTemplate):
    """
    DDF00127: An encounter must only be scheduled at a timing that is defined within the same study design as the encounter.

    Applies to: Encounter
    Attributes: scheduledAt
    """

    def __init__(self):
        super().__init__(
            "DDF00127",
            RuleTemplate.ERROR,
            "An encounter must only be scheduled at a timing that is defined within the same study design as the encounter.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00063(RuleTemplate):
    """
    DDF00063: A standard code alias is not expected to be equal to the standard code (e.g. no equal code or decode for the same coding system version is expected).

    Applies to: AliasCode
    Attributes: standardCodeAliases
    """

    def __init__(self):
        super().__init__(
            "DDF00063",
            RuleTemplate.ERROR,
            "A standard code alias is not expected to be equal to the standard code (e.g. no equal code or decode for the same coding system version is expected).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0193(RuleTemplate):
    """
    CHK0193: There must be exactly one sponsor study identifier (i.e., a study identifier whose scope is an organization that is identified as the organization for the sponsor study role).

    Applies to: StudyIdentifier
    Attributes: scope
    """

    def __init__(self):
        super().__init__(
            "CHK0193",
            RuleTemplate.ERROR,
            "There must be exactly one sponsor study identifier (i.e., a study identifier whose scope is an organization that is identified as the organization for the sponsor study role).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import enum
import datetime
from uuid import UUID


# Example, see https://stackoverflow.com/questions/10252010/serializing-class-instance-to-json
def serialize_as_json(obj):
    if isinstance(obj, enum.Enum):
        return obj.value
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    else:
        return obj.__dict__



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00046(RuleTemplate):
    """
    DDF00046: A timing must only be specified as being relative to/from a scheduled activity/decision instance that is defined within the same timeline as the timing.

    Applies to: Timing
    Attributes: relativeFromScheduledInstance, relativeToScheduledInstance
    """

    def __init__(self):
        super().__init__(
            "DDF00046",
            RuleTemplate.ERROR,
            "A timing must only be specified as being relative to/from a scheduled activity/decision instance that is defined within the same timeline as the timing.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00032(RuleTemplate):
    """
    DDF00032: Within a study version, if more than 1 business therapeutic area is defined then they must be distinct.

    Applies to: StudyVersion
    Attributes: businessTherapeuticAreas
    """

    def __init__(self):
        super().__init__(
            "DDF00032",
            RuleTemplate.ERROR,
            "Within a study version, if more than 1 business therapeutic area is defined then they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0237(RuleTemplate):
    """
    CHK0237: A masking is expected to be defined for at least one study role in a study design with a blinding schema that is not open label or double blind.

    Applies to: StudyRole
    Attributes: masking
    """

    def __init__(self):
        super().__init__(
            "CHK0237",
            RuleTemplate.ERROR,
            "A masking is expected to be defined for at least one study role in a study design with a blinding schema that is not open label or double blind.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code


class StudySite(ApiBaseModelWithIdNameLabelAndDesc):
    country: Code
    instanceType: Literal["StudySite"]



from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.builder.builder import Builder
from usdm4.api.population_definition import StudyDesignPopulation
from usdm4.api.eligibility_criterion import (
    EligibilityCriterion,
    EligibilityCriterionItem,
)


class PopulationAssembler(BaseAssembler):
    """
    Assembler responsible for processing population-related data and creating StudyDesignPopulation objects.

    This assembler handles the creation of study population definitions, including population criteria,
    cohort definitions, and subject enrollment information. It processes population data from the
    input structure and creates the appropriate USDM population objects.
    """

    MODULE = "usdm4.assembler.population_assembler.PopulationAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the PopulationAssembler.

        Args:
            builder (Builder): The builder instance for creating USDM objects
            errors (Errors): Error handling instance for logging issues
        """
        super().__init__(builder, errors)
        self.clear()

    def clear(self):
        self._population = None
        self._cohorts = []
        self._ec_items = []
        self._eci_items = []

    def execute(self, data: dict) -> None:
        """
        Processes population data and creates a StudyDesignPopulation object.

        Args:
            data (dict): A dictionary containing population definition data.
                        The data parameter must have the following structure:

                        {
                            "label": str,              # Human-readable label for the population
                            "inclusion_exclusion: {
                                "inclusion": list[str],
                                "exclusion": list[str],
                            }
                            # Additional optional fields may be includes in the future:
                            # "criteria": list,         # List of eligibility criteria
                            # "cohorts": list,          # List of population cohorts/subgroups
                            # "enrollment": dict,       # Subject enrollment information
                            # "analysis_populations": list  # Analysis population definitions
                        }

                        Required fields:
                        - "label": A string that provides a human-readable name for the population.
                          This will be used to generate both the display label and the internal name
                          (converted to uppercase with spaces replaced by hyphens).

                        The current implementation creates a basic population definition with:
                        - name: Generated from label (uppercase, spaces -> hyphens)
                        - label: Direct copy of the input label
                        - description: Default placeholder text
                        - includesHealthySubjects: Default to True
                        - criteria: Empty list (to be populated by future enhancements)

        Returns:
            None: The created population is stored in self._population property

        Raises:
            Exception: If population creation fails, logged via error handler
        """
        try:
            if data:
                self._ie(data["inclusion_exclusion"])

                # Extract required label field and create population parameters
                # The label is used for both display purposes and name generation
                params = {
                    "name": data["label"]
                    .upper()
                    .replace(" ", "-"),  # Convert label to internal name format
                    "label": data["label"],  # Keep original label for display
                    "description": "The study population, currently blank",  # Default description
                    "includesHealthySubjects": True,  # Default assumption
                    "criteria": self._ec_items,
                }

                # Create the StudyDesignPopulation object using the builder
                self._population = self._builder.create(StudyDesignPopulation, params)
            else:
                self._errors.info(
                    "No population to build, no data",
                    KlassMethodLocation(self.MODULE, "execute"),
                )
        except Exception as e:
            self._errors.exception(
                "Failed during creation of population",
                e,
                KlassMethodLocation(self.MODULE, "execute"),
            )

    @property
    def population(self) -> StudyDesignPopulation:
        return self._population

    @property
    def criteria_items(self) -> list[EligibilityCriterionItem]:
        return self._eci_items

    def _ie(self, criteria: dict) -> None:
        self._collection(
            criteria["inclusion"], "C25532", "INCLUSION", "INC", "Inclusion"
        )
        self._collection(
            criteria["exclusion"], "C25370", "EXCLUSION", "EXC", "Exclusion"
        )

    def _collection(
        self, criteria: list[str], code: str, decode: str, prefix: str, label: str
    ) -> None:
        for index, text in enumerate(criteria):
            try:
                category = self._builder.cdisc_code(code, decode)
                params = {
                    "name": f"{prefix}-I{index + 1}",
                    "label": f"{label} item {index + 1} ",
                    "description": "",
                    "text": text,
                }
                eci_item = self._builder.create(EligibilityCriterionItem, params)
                self._eci_items.append(eci_item)
                params = {
                    "name": f"{prefix}{index + 1}",
                    "label": f"{label} criterion {index + 1} ",
                    "description": "",
                    "criterionItemId": eci_item.id,
                    "category": category,
                    "identifier": f"{index + 1}",
                }
                self._ec_items.append(
                    self._builder.create(EligibilityCriterion, params)
                )
            except Exception as e:
                location = KlassMethodLocation(self.MODULE, "_collection")
                self._errors.exception(
                    f"Failed during creation of criterion '{text}", e, location
                )



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00151(RuleTemplate):
    """
    DDF00151: If geographic scope type is global then there must be only one geographic scope specified.

    Applies to: GovernanceDate
    Attributes: geographicScopes
    """

    def __init__(self):
        super().__init__(
            "DDF00151",
            RuleTemplate.ERROR,
            "If geographic scope type is global then there must be only one geographic scope specified.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import datetime
import dateutil.parser as parser
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.builder.builder import Builder
from usdm4.api.alias_code import AliasCode
from usdm4.api.code import Code


class Encoder:
    MODULE = "usdm4.encoder.encoder.Encoder"

    ZERO_DURATION = "PT0M"

    PHASE_MAP = [
        (
            ["0", "PRE-CLINICAL", "PRE CLINICAL"],
            {"code": "C54721", "decode": "Phase 0 Trial"},
        ),
        (["1", "I"], {"code": "C15600", "decode": "Phase I Trial"}),
        (["1-2"], {"code": "C15693", "decode": "Phase I/II Trial"}),
        (["1/2"], {"code": "C15693", "decode": "Phase I/II Trial"}),
        (["1/2/3"], {"code": "C198366", "decode": "Phase I/II/III Trial"}),
        (["1/3"], {"code": "C198367", "decode": "Phase I/III Trial"}),
        (["1A", "IA"], {"code": "C199990", "decode": "Phase Ia Trial"}),
        (["1B", "IB"], {"code": "C199989", "decode": "Phase Ib Trial"}),
        (["2", "II"], {"code": "C15601", "decode": "Phase II Trial"}),
        (["2-3", "II-III"], {"code": "C15694", "decode": "Phase II/III Trial"}),
        (["2A", "IIA"], {"code": "C49686", "decode": "Phase IIa Trial"}),
        (["2B", "IIB"], {"code": "C49688", "decode": "Phase IIb Trial"}),
        (["3", "III"], {"code": "C15602", "decode": "Phase III Trial"}),
        (["3A", "IIIA"], {"code": "C49687", "decode": "Phase IIIa Trial"}),
        (["3B", "IIIB"], {"code": "C49689", "decode": "Phase IIIb Trial"}),
        (["4", "IV"], {"code": "C15603", "decode": "Phase IV Trial"}),
        (["5", "V"], {"code": "C47865", "decode": "Phase V Trial"}),
    ]
    STATUS_MAP = [
        (["APPROVED"], {"code": "C25425", "decode": "Approved"}),
        (["DRAFT", "DFT"], {"code": "C85255", "decode": "Draft"}),
        (["FINAL"], {"code": "C25508", "decode": "Final"}),
        (["OBSOLETE"], {"code": "C63553", "decode": "Obsolete"}),
        (
            ["PENDING", "PRENDING REVIEW"],
            {"code": "C188862", "decode": "Pending Review"},
        ),
    ]
    REASON_MAP = [
        {"code": "C207612", "decode": "Regulatory Agency Request To Amend"},
        {"code": "C207608", "decode": "New Regulatory Guidance"},
        {"code": "C207605", "decode": "IRB/IEC Feedback"},
        {"code": "C207609", "decode": "New Safety Information Available"},
        {"code": "C207606", "decode": "Manufacturing Change"},
        {"code": "C207602", "decode": "IMP Addition"},
        {"code": "C207601", "decode": "Change In Strategy"},
        {"code": "C207600", "decode": "Change In Standard Of Care"},
        {
            "code": "C207607",
            "decode": "New Data Available (Other Than Safety Data)",
        },
        {"code": "C207604", "decode": "Investigator/Site Feedback"},
        {"code": "C207611", "decode": "Recruitment Difficulty"},
        {
            "code": "C207603",
            "decode": "Inconsistency And/Or Error In The Protocol",
        },
        {"code": "C207610", "decode": "Protocol Design Error"},
        {"code": "C17649", "decode": "Other"},
        {"code": "C48660", "decode": "Not Applicable"},
    ]

    def __init__(self, builder: Builder, errors: Errors):
        self._builder: Builder = builder
        self._errors: Errors = errors

    def phase(self, text: str) -> AliasCode:
        phase = text
        for word in ["PHASE", "TRIAL"]:
            phase = phase.upper().replace(word, "").strip() if phase else ""
        for tuple in self.PHASE_MAP:
            if phase in tuple[0]:
                entry = tuple[1]
                cdisc_phase_code = self._builder.cdisc_code(
                    entry["code"],
                    entry["decode"],
                )
                self._errors.info(
                    f"Trial phase '{phase}' decoded as '{entry['code']}', '{entry['decode']}'",
                    location=KlassMethodLocation(self.MODULE, "phase"),
                )
                return self._builder.alias_code(cdisc_phase_code)
        cdisc_phase_code = self._builder.cdisc_code(
            "C48660",
            "[Trial Phase] Not Applicable",
        )
        self._errors.warning(
            f"Trial phase '{phase}' not decoded",
            location=KlassMethodLocation(self.MODULE, "phase"),
        )
        return self._builder.alias_code(cdisc_phase_code)

    def document_status(self, text: str) -> Code:
        status = text.upper().strip() if text else ""
        for tuple in self.STATUS_MAP:
            if status in tuple[0]:
                entry = tuple[1]
                cdisc_code = self._builder.cdisc_code(
                    entry["code"],
                    entry["decode"],
                )
                self._errors.info(
                    f"Document status '{status}' decoded as '{entry['code']}', '{entry['decode']}'",
                    location=KlassMethodLocation(self.MODULE, "document_status"),
                )
                return cdisc_code
        cdisc_code = self._builder.cdisc_code("C85255", "Draft")
        self._errors.warning(
            f"Document status '{status}' not decoded",
            location=KlassMethodLocation(self.MODULE, "document_status"),
        )
        return cdisc_code

    def amendment_reason(self, reason_str: str):
        if reason_str:
            parts = reason_str.split(":")
            # print(f"PARTS: {parts}")
            if len(parts) >= 2:
                reason_text = parts[1]
                # print(f"REASON: {reason_text}")
                for reason in self.REASON_MAP:
                    if reason_text in reason["decode"]:
                        self._errors.info(
                            f"Amendment reason '{reason_text}' decoded as '{reason['code']}', '{reason['decode']}'"
                        )
                        code = self._builder.cdisc_code(
                            reason["code"], reason["decode"]
                        )
                        return {"code": code, "other_reason": ""}
            self._errors.warning(
                f"Unable to decode amendment reason '{reason_str}'",
                location=KlassMethodLocation(self.MODULE, "amendment_reason"),
            )
            code = self._builder.cdisc_code("C17649", "Other")
            return {"code": code, "other_reason": parts[-1].strip()}
        self._errors.warning(
            "Amendment reason not decoded, missing text",
            location=KlassMethodLocation(self.MODULE, "amendment_reason"),
        )
        code = self._builder.cdisc_code("C17649", "Other")
        return {"code": code, "other_reason": "No reason text found"}

    def to_date(self, text: str) -> datetime.datetime | None:
        try:
            input_text = text.strip()
            if input_text:
                return parser.parse(input_text)
            else:
                return None
        except Exception as e:
            self._errors.exception(
                f"Failed to decode date text '{text}'",
                e,
                location=KlassMethodLocation(self.MODULE, "to_date"),
            )
            return None

    def iso8601_duration(self, value: int, unit: str) -> str:
        try:
            unit_text: str = unit.strip()
            if unit_text.upper() in ["Y", "YRS", "YR", "YEARS", "YEAR"]:
                return f"P{value}Y"
            if unit_text.upper() in ["MTHS", "MTH", "MONTHS", "MONTH"]:
                return f"P{value}M"
            if unit_text.upper() in ["W", "WKS", "WK", "WEEKS", "WEEK"]:
                return f"P{value}W"
            if unit_text.upper() in ["D", "DYS", "DY", "DAYS", "DAY"]:
                return f"P{value}D"
            if unit_text.upper() in ["H", "HRS", "HR", "HOURS", "HOUR"]:
                return f"PT{value}H"
            if unit_text.upper() in ["M", "MINS", "MIN", "MINUTES", "MINUTE"]:
                return f"PT{value}M"
            if unit_text.upper() in ["S", "SECS", "SEC", "SECONDS", "SECOND"]:
                return f"PT{value}S"
            self._errors.warning(
                f"Failed to encode ISO8601 duration of '{value}, {unit}'"
            )
            return self.ZERO_DURATION
        except Exception as e:
            self._errors.exception(
                f"Failed to encode ISO8601 duration of '{value}, {unit}'",
                e,
                location=KlassMethodLocation(self.MODULE, "iso8601_duration"),
            )
            return self.ZERO_DURATION



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00027(RuleTemplate):
    """
    DDF00027: To ensure consistent ordering, the same instance must not be referenced more than once as previous or next.

    Applies to: Activity, EligibilityCriterion, Encounter, NarrativeContent, StudyEpoch
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00027",
            RuleTemplate.ERROR,
            "To ensure consistent ordering, the same instance must not be referenced more than once as previous or next.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithId


class PersonName(ApiBaseModelWithId):
    text: Union[str, None] = None
    familyName: Union[str, None] = None
    givenNames: List[str] = []
    prefixes: List[str] = []
    suffixes: List[str] = []
    instanceType: Literal["PersonName"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00050(RuleTemplate):
    """
    DDF00050: A study arm must only reference study populations or cohorts that are defined within the same study design as the study arm.

    Applies to: StudyArm
    Attributes: populations
    """

    def __init__(self):
        super().__init__(
            "DDF00050",
            RuleTemplate.ERROR,
            "A study arm must only reference study populations or cohorts that are defined within the same study design as the study arm.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00139(RuleTemplate):
    """
    DDF00139: An identified organization is not expected to have more than one identifier for the study.

    Applies to: StudyIdentifier
    Attributes: studyIdentifierScope
    """

    def __init__(self):
        super().__init__(
            "DDF00139",
            RuleTemplate.ERROR,
            "An identified organization is not expected to have more than one identifier for the study.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .identifier import MedicalDeviceIdentifier
from .comment_annotation import CommentAnnotation


class MedicalDevice(ApiBaseModelWithIdNameLabelAndDesc):
    hardwareVersion: Union[str, None] = None
    softwareVersion: Union[str, None] = None
    embeddedProductId: Union[str, None] = None
    sourcing: Union[Code, None] = None
    identifiers: List[MedicalDeviceIdentifier] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["MedicalDevice"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0173(RuleTemplate):
    """
    CHK0173: Each defined elibility criterion must be used by at least one study population or cohort.

    Applies to: StudyVersion
    Attributes: criteria
    """

    def __init__(self):
        super().__init__(
            "CHK0173",
            RuleTemplate.ERROR,
            "Each defined elibility criterion must be used by at least one study population or cohort.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0182(RuleTemplate):
    """
    CHK0182: Narrative content is expected to point to a child and/or to a content item text.

    Applies to: NarrativeContent
    Attributes: children, contentItem
    """

    def __init__(self):
        super().__init__(
            "CHK0182",
            RuleTemplate.ERROR,
            "Narrative content is expected to point to a child and/or to a content item text.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00028(RuleTemplate):
    """
    DDF00028: An activity must only reference activities that are specified within the same study design.

    Applies to: Activity
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00028",
            RuleTemplate.ERROR,
            "An activity must only reference activities that are specified within the same study design.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0253(RuleTemplate):
    """
    CHK0253: A medical device sourcing must be specified using the sourcing (Cxxx) DDF codelist.

    Applies to: MedicalDevice
    Attributes: sourcing
    """

    def __init__(self):
        super().__init__(
            "CHK0253",
            RuleTemplate.ERROR,
            "A medical device sourcing must be specified using the sourcing (Cxxx) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from pydantic import Field
from .api_base_model import ApiBaseModel
from .study_definition_document import StudyDefinitionDocument
from .study_version import StudyVersion
from uuid import UUID


class Study(ApiBaseModel):
    id: Union[UUID, None] = None
    name: str = Field(min_length=1)
    description: Union[str, None] = None
    label: Union[str, None] = None
    versions: List[StudyVersion] = []
    documentedBy: List[StudyDefinitionDocument] = []
    instanceType: Literal["Study"]

    def document_by_template_name(self, name: str) -> StudyDefinitionDocument | None:
        return next(
            (x for x in self.documentedBy if x.templateName.upper() == name.upper()),
            None,
        )

    def first_version(self) -> StudyVersion | None:
        try:
            return self.versions[0]
        except Exception:
            return None
